/*
	Плагин: Bypass Guard
	Основное назначение: Противодействие обходу банов по SteamID/IP
	Использующийся инструментарий:
		* Агрегатор номера автономной системы (номера провайдера), к которой принадлежит IP (Whois-модуль)
		* Агрегатор кода страны (GeoIP-модуль)
		* Агрегатор статуса IP-адреса в онлайн-сервисе проверок на Proxy/VPN [ Curl-модуль + сервис https://iphub.info/ ]
		* Чёрный и белый списки IP-адресов

	Данный плагин частично основан на функционале двух других плагинов:
		* Proxy/VPN check 1.1b от juice: https://dev-cs.ru/resources/475/
		* SubnetBan 2.2 от Lev: https://dev-cs.ru/resources/162/

	Требования:
		* Amx Mod X 1.8.3, либо новее
		* Reapi
		* Свежая geo-база (country): http://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz
		* AGHL.RU Dev Team Whois Module: http://aghl.ru/forum/viewtopic.php?f=19&p=24798
		* Amxx Curl: https://github.com/Polarhigh/AmxxCurl/releases

	Доступные команды:
		* bg_allow_steamid <steamid> - Выдаёт указанному steamid иммунитет ко всем проверкам

		Внимание! Внесение AS/IP в чёрный список не проверяет сервер на наличие игроков, подпадающих под добавленное ограничение.
		Другими словами, забанив AS/IP игрока, который в данный момент находится на сервере, вы должны кикнуть его самостоятельно.

		* bg_as_blacklist_add <as number> "<comment>"
		* bg_as_blacklist_del <as number>
		* bg_as_blacklist_show <page>

		* bg_as_whitelist_add <as number> "<comment>"
		* bg_as_whitelist_del <as number>
		* bg_as_whitelist_show <page>

		* bg_ip_blacklist_add <start ip> <end ip> "<comment>"
		* bg_ip_blacklist_del <start ip> <end ip>
		* bg_ip_blacklist_show <page>

		* bg_ip_whitelist_add <start ip> <end ip> "<comment>"
		* bg_ip_whitelist_del <start ip> <end ip>
		* bg_ip_whitelist_show <page>

		* bg_find_ip <ip> - Позволяет проверить наличие указанного IP в диапазонах чёрного и белого списков

		* bg_get_as_by_ip <ip> - Позволяет получить AS-номер для указанного IP

		* bg_find_as <as number> - Позволяет проверить наличие указанного AS-номера в чёрном и белом списках

		* bg_check_ip <ip> - Позволяет проверить IP-адрес на причастность к Proxy/VPN (минуя весь кеш)

		* bg_flush_data <1-5>:
			1 - Обнулить хранилище, содержащее результаты проверок IP-адресов на Proxy/VPN (sql + trie)
			2 - Обнулить хранилище, содержащее steamid'ы, имеющие иммунитет ко всем проверкам (nvault)
			3 - Обнулить кеш проверок AS-номеров (trie, короткий кеш)
			4 - Обнулить чёрный и белый список IP, удалить файл-список диапазонов IP
			5 - Обнулить чёрный и белый список AS-номеров, кеш проверок AS-номеров, удалить файл-список AS-номеров

		* bg_status - Выводит информацию о присутствующих игроках
*/

/* Thx to:
	fantom за пример работы с форматом JSON
	juice за плагин 'Proxy/VPN check'
	Lev за плагин 'SubnetBan' и модуль 'Whois'
	Polarhigh за модуль 'Amxx Curl'
*/

/* История обновлений:
	08.01.2019:
		* Закрытый релиз
	09.01.2019:
		* Добавлен белый список AS
		* Доработки функционала консольных команд
		* Улучшена информативность лог-файла
		* Фиксы обнаруженных ошибок
	19.01.2019:
		* Смена сервиса проверки на Proxy/VPN с 'mind-media.com' на 'iphub.info'
		* Смена метода работы с WEB (HTTP:X -> Amxx Curl)
		* Теперь использование команд (добавление/удаление) логируется в файл
		* Логирование разделёно на несколько частей, см. 'LOG_ENUM'
		* Добавлены квары, предоставляющие некоторую гибкость поведения
		* Команда 'bg_get_as_by_ip' теперь так же показывает поле 'Description' (провайдер)
		* Добавлена команда 'bg_status'
	22.01.2019:
		* Улучшено поведение при возникновении ошибок при проверке на Proxy/VPN
		* Для команды 'bg_allow_steamid' добавлен учёт 'VALVE_'
		* Список разрешённых стран переведён на загрузку из файла
		* Загрузка API-ключей переведена на загрузку из файла
		* Информация, выводящаяся игроку при кике, выведена в lang-файл
		* Добавлен квар 'bypass_guard_show_url'
		* Добавлен квар 'bypass_guard_check_delay'
		* Добавлен квар 'bypass_guard_kick_delay'
	24.01.2019:
		* Убран код, пропускающий казахов без проверки на Proxy/VPN (проблема была актуальна для mind-media.com)
		* Устранение ряда незначительных недосмотров
		* Все команды: srvcmd -> concmd, логирование использования сопровождается информацией о пользователе
		* Открытый релиз
	24.02.2019:
		* Добавлен учёт отсутствия квара 'amx_default_access', спасибо w0w
		* Пофикшен баг с пустым выводом команды 'bg_status', спасибо Nordic Warrior
		* Хранение статуса IP-адресов переведено на SQL (не забудьте прописать sqlite или mysql в 'configs/modules.ini'). Спасибо fantom.
		* Добавлены квары 'bg_sql_'. Рекомендуется удалить конфиг в 'configs/plugins', дабы он пересоздался.
		* Добавлен квар 'bg_divide_logs_by_month'
		* Для команды 'bg_get_as_by_ip' добавлен вывод страны
		* Добавлена команда 'bg_check_ip', позволяющая проверить IP-адрес на причастность к Proxy/VPN
		* Улучшена информативность поля "тип доступа" для команды 'bg_status', логов 'ALLOW.log', 'DENY.log', и 'PROXY_CHECK.log'
		* Исправление выявленных недочётов в логике работы проверок
		* Отдельная благодарность за подсказки: fl0wer, Sonyx, bionext
*/

new const PLUGIN_DATE[] = "24.02.2019"

/* ----------------------- */

//#define DEBUG // Режим отладки. В рабочей версии должен быть отключён.

#define AUTO_CFG // Создавать конфиг с кварами в 'configs/plugins', и запускать его ?

#define ACCESS_FLAG ADMIN_CFG // Флаг доступа по-умолчанию ко всем консольным командам

const MAX_KEYS = 5 // Макс. кол-во API-ключей. Увеличить при необходимости.

const MAX_CODES = 32 // Макс. кол-во разрешённых стран. Увеличить при необходимости.

new const DIR_NAME[] = "bypass_guard" // 'configs/%s', 'data/%s', 'logs/%s'
new const IP_FILE_NAME[] = "ip_list.ini"
new const AS_FILE_NAME[] = "as_list.ini"
new const COUNTRY_FILE_NAME[] = "allowed_countries.ini"
new const KEY_FILE_NAME[] = "api_keys.ini"
new const IMMUNE_STEAMS_VAULT[] = "bypass_guard_steams"
stock const DEBUG_VAULT[] = "bg_debug"

enum _:LOG_ENUM { // Don't touch this!
	LOG__PROXY_CHECK,
	LOG__CMD,
	LOG__ERROR,
	//
	// <--- new log type goes here!
	//
	LOG__LONG_QUERY, // Must be the last before LOG__DENY
	// Must be the last in the enum --->
	LOG__DENY,
	LOG__ALLOW
	// <---
}

new const LOG_NAME[LOG_ENUM][] = {
	"PROXY_CHECK.log",
	"CMD.log",
	"ERROR.log",
	"LONG_QUERY.log",
	"DENY.log",
	"ALLOW.log"
}

/* ----------------------- */

#include <amxmodx>
#include <geoip>
#include <nvault>
#include <sqlx>
#include <reapi>
#include <whois>
#include <curl>
#include <json>

#define chx charsmax

#define IP_LEN 16
#define AUTHID_LEN 24
#define COUNTRY_LEN 64
#define COMMENT_LEN 48
#define AS_LEN 16
#define CURL_BUFFER_SIZE 512
#define DESC_LEN 64
#define CODE_LEN 3
#define ACCESS_LEN 32
#define API_KEY_LEN 64
#define SQL_ACC_LEN 64
#define QUERY_LEN 1472

new const _NA_[] = "N/A"

enum ( += MAX_PLAYERS + 1 ) {
	TASKID__RETRY_INIT_QUERY = 100
}

/* --- */

enum {
	QUERY__INIT_SYSTEM,
	QUERY__CHECK_IP,
	QUERY__FLUSH_IPS,
	// Request without answer waiting --->
	QUERY__INSERT_IP
}

enum _:SQL_DATA_STRUCT {
	SQL_DATA__QUERY_TYPE,
	SQL_DATA__PLAYER_INDEX,
	SQL_DATA__PLAYER_USERID
}

/* --- */

enum _:SQL_TABLE_COLUMN_ENUM {
	SQL_COLUMN__ID,
	SQL_COLUMN__IP,
	SQL_COLUMN__IP_STATUS,
	SQL_COLUMN__TIMESTAMP
}

new const SQL_TABLE_COLUMNS[SQL_TABLE_COLUMN_ENUM][] = {
	"id",
	"ip",
	"ip_status",
	"timestamp"
}

/* --- */

enum _:STATES_ENUM {
	STATE__NO,
	STATE__NORMAL,
	STATE__SUSPICIOUS,
	STATE__BAN,
	STATE__WHITELIST
}

new const IP_STATES[STATES_ENUM][] = {
	"",
	"Normal",
	"Suspicious",
	"Proxy/VPN",
	""
}

enum KICK_TYPE_ENUM {
	KICK_TYPE__AS,
	KICK_TYPE__IP_BAN,
	KICK_TYPE__COUNTRY,
	KICK_TYPE__PROXY_CHECK,
	KICK_TYPE__PROXY_SQL,
	KICK_TYPE__PROXY_CACHE,
	KICK_TYPE__SUSPICIOUS_CHECK,
	KICK_TYPE__SUSPICIOUS_SQL,
	KICK_TYPE__SUSPICIOUS_CACHE,
	KICK_TYPE__AS_CHECK_FAIL,
	KICK_TYPE__PROXY_CHECK_FAIL
}

enum {
	WHOIS_QUERY_TYPE__AUTO,
	WHOIS_QUERY_TYPE__MANUAL
}

enum _:LIST_TYPE_ENUM {
	LIST_TYPE__BLACKLIST,
	LIST_TYPE__WHITELIST
}

new const LIST_NAME[LIST_TYPE_ENUM][] = { "blacklist", "whitelist" }

enum _:RANGE_DATA_STRUCT {
	RDS__START_IP,
	RDS__END_IP,
	RDS__COMMENT[COMMENT_LEN]
}

enum _:AS_DATA_STRUCT {
	ASD__NUMBER[AS_LEN],
	ASD__COMMENT[COMMENT_LEN]
}

enum _:AS_TRIE_STRUCT {
	AST__STATE,
	AST__ARRAY_POS
}

enum _:WHOIS_DATA_STRUCT {
	WHOIS_DATA__QUERY_TYPE,
	WHOIS_DATA__QUERY_TRIES,
	WHOIS_DATA__PLAYER_USERID
}

enum _:CURL_DATA_STRUCT {
	CURL_DATA__FILE_HANDLE,
	CURL_DATA__SLIST_HANDLE,
	CURL_DATA__PLAYER_INDEX,
	CURL_DATA__PLAYER_USERID
}

enum /*IPHUB_CHECK_STATE*/ {
	IP_CHECK_STATE__NORMAL,
	IP_CHECK_STATE__DETECT,
	IP_CHECK_STATE__SUSPICIOUS
}

enum _:AGREGATE_TYPE_ENUM {
	AGREGATE_IP__CACHE,
	AGREGATE_IP__SQL
}

enum _:EXPECTED_JSON_KEY_COUNT {
	JSON_KEY__IP,
	JSON_KEY__COUNTRY_CODE,
	JSON_KEY__COUNTRY_NAME,
	JSON_KEY__AS,
	JSON_KEY__PROVIDER,
	JSON_KEY__CHECK_STATE,
	JSON_KEY__HOSTNAME
}

enum _:PCVAR_ENUM {
	PCVAR__SQL_MODE,
	PCVAR__HOST,
	PCVAR__USER,
	PCVAR__PWD,
	PCVAR__DB,
	PCVAR__TABLE,
	PCVAR__CONN_TIMEOUT,
	PCVAR__CREATE_TABLE,
	PCVAR__PRUNE_OLD_IPS,
	PCVAR__DIVIDE_LOGS
}

enum _:CVAR_ENUM {
	CVAR__PLUGIN_ENABLED,
	CVAR__BAN_SUSPICIOUS,
	CVAR__MAX_WHOIS_QUERY_TRIES,
	CVAR__KICK_IF_CANT_CHECK,
	CVAR__ALLOW_STEAM,
	CVAR__SHOW_URL,
	CVAR__CHECK_DELAY,
	CVAR__KICK_DELAY,
	Float:CVAR_F__REQUEST_TIMEOUT,
	Float:CVAR_F__LONG_QUERY_TIME
}

enum _:IP_INFO_STRUCT {
	IP_INFO__AS[AS_LEN],
	IP_INFO__DESC[DESC_LEN]
}

new g_eIpInfoData[IP_INFO_STRUCT], g_iManualWhoisState
new Array:g_eIpArray[LIST_TYPE_ENUM], g_eRangeData[RANGE_DATA_STRUCT], g_iIpCount[LIST_TYPE_ENUM]
new g_hImmunity = INVALID_HANDLE, g_szIP[IP_LEN]
new g_szFile[PLATFORM_MAX_PATH], g_szDataDir[PLATFORM_MAX_PATH]
new g_szAuthID[AUTHID_LEN], g_szRequest[96], g_szIpListFile[PLATFORM_MAX_PATH]
new Trie:g_tAsCache, Trie:g_tAsNumbers, g_szAsListFile[PLATFORM_MAX_PATH]
new g_eAsData[AS_DATA_STRUCT], g_iAsCount[LIST_TYPE_ENUM], Array:g_aAsArray[LIST_TYPE_ENUM]
new g_iDefAccFlags, g_szGetWhoisIP[IP_LEN], g_eAsTrieData[AS_TRIE_STRUCT]
new Trie:g_tCurl, bool:g_bPluginEnded, g_eCvar[CVAR_ENUM]
new g_eLogFile[LOG_ENUM][PLATFORM_MAX_PATH], Trie:g_tIpInfo
new g_szAccess[MAX_PLAYERS + 1][ACCESS_LEN]
new g_szAsNumber[MAX_PLAYERS + 1][AS_LEN]
new g_szDesc[MAX_PLAYERS + 1][DESC_LEN]
new g_szCode[MAX_PLAYERS + 1][CODE_LEN * 2]
new g_szCountry[MAX_PLAYERS + 1][COUNTRY_LEN]
new g_szAllowedCodes[MAX_CODES][CODE_LEN]
new g_iCodesCount, g_iCurrentKey, g_iLoadedKeys, g_szApiKey[MAX_KEYS][API_KEY_LEN]
new g_pCvar[PCVAR_ENUM]
new Trie:g_tIpProxyStatus
new Trie:g_tIpMatch
new bool:g_bSystemAvailable
new Handle:g_hSqlTuple
new g_eSqlData[SQL_DATA_STRUCT]
new g_szQuery[QUERY_LEN]
new g_szTableName[64]
new g_szSqlMode[10]
new g_szGetProxyIP[IP_LEN]
stock g_hDebug

/* -------------------- */

public plugin_init() {
	register_plugin("Bypass Guard", PLUGIN_DATE, "mx?!")
	register_dictionary("bypass_guard.txt")

	g_pCvar[PCVAR__SQL_MODE] = create_cvar( "bg_sql_mode", "sqlite",
		.description = "sqlite - use SQLite; mysql - use MySQL; don't forget to enable module in 'modules.ini'" );

	g_pCvar[PCVAR__HOST] = create_cvar("bg_sql_host", "127.0.0.1", FCVAR_PROTECTED)
	g_pCvar[PCVAR__USER] = create_cvar("bg_sql_user", "", FCVAR_PROTECTED)
	g_pCvar[PCVAR__PWD] = create_cvar("bg_sql_pwd", "", FCVAR_PROTECTED)
	g_pCvar[PCVAR__DB] = create_cvar("bg_sql_db", "bypass_guard", FCVAR_PROTECTED)
	g_pCvar[PCVAR__TABLE] = create_cvar("bg_sql_table", "bypass_guard_ips", FCVAR_PROTECTED)
	g_pCvar[PCVAR__CONN_TIMEOUT] = create_cvar("bg_sql_conn_timeout", "15", .has_min = true, .min_val = 1.0)
	g_pCvar[PCVAR__CREATE_TABLE] = create_cvar("bg_sql_create_table", "1", .description = "Create table automaticaly?")

	g_pCvar[PCVAR__PRUNE_OLD_IPS] = create_cvar( "bg_sql_prune_old_ips", "30",
		.description = "Rows older than # days will be removed from DB (0 = disable)" );

	bind_pcvar_float( create_cvar("bg_sql_request_timeout", "10.0", .has_min = true, .min_val = 2.0),
		g_eCvar[CVAR_F__REQUEST_TIMEOUT] );

	bind_pcvar_float(create_cvar("bg_sql_long_query_time", "10.0"), g_eCvar[CVAR_F__LONG_QUERY_TIME])

	bind_pcvar_num( create_cvar("bypass_guard_enabled", "1",
		.description = "Enable/Disable plugin work"), g_eCvar[CVAR__PLUGIN_ENABLED] );

	bind_pcvar_num( create_cvar("bypass_guard_ban_suspicious", "0",
		.description = "Ban suspicious IPs?"), g_eCvar[CVAR__BAN_SUSPICIOUS] );

	bind_pcvar_num( create_cvar("bypass_guard_kick_if_cant_check", "0",
		.description = "Kick player if check fails?"), g_eCvar[CVAR__KICK_IF_CANT_CHECK] );

	bind_pcvar_num( create_cvar("bypass_guard_allow_steam", "1",
		.description = "Allow steam players without checks?"), g_eCvar[CVAR__ALLOW_STEAM] );

	bind_pcvar_num( create_cvar("bypass_guard_show_url", "1",
		.description = "Show URL when player gets kicked?"), g_eCvar[CVAR__SHOW_URL] );

	bind_pcvar_num( create_cvar("bypass_guard_max_whois_tries", "2", .has_min = true, .min_val = 1.0,
		.description = "Max whois query tries"), g_eCvar[CVAR__MAX_WHOIS_QUERY_TRIES] );

	bind_pcvar_float( create_cvar("bypass_guard_check_delay", "0.1", .has_min = true, .min_val = 0.1,
		.description = "Player check delay"), Float:g_eCvar[CVAR__CHECK_DELAY] );

	bind_pcvar_float( create_cvar("bypass_guard_kick_delay", "1.0", .has_min = true, .min_val = 0.1,
		.description = "Player kick delay (low values may break URL showing!)"), Float:g_eCvar[CVAR__KICK_DELAY] );

	g_pCvar[PCVAR__DIVIDE_LOGS] = create_cvar( "bg_divide_logs_by_month", "1",
		.description = "Divide 'ALLOW' and 'DENY' logs by month?" );

#if defined AUTO_CFG
	AutoExecConfig()
#endif
}

/* -------------------- */

public plugin_cfg() {
	g_hImmunity = nvault_open(IMMUNE_STEAMS_VAULT)

	if(g_hImmunity == INVALID_HANDLE) {
		set_fail_state("[Immunity] Error opening nVault!")
	}

#if defined DEBUG
	g_hDebug = nvault_open(DEBUG_VAULT)
#endif

	/* --- */

	get_localinfo("amxx_datadir", g_szDataDir, chx(g_szDataDir))
	format(g_szDataDir, chx(g_szDataDir), "%s/%s", g_szDataDir, DIR_NAME)

	if(!dir_exists(g_szDataDir)) {
		mkdir(g_szDataDir)
	}

	/* --- */

	register_concmd("bg_allow_steamid", "concmd_AllowSteamID", ACCESS_FLAG)
	register_concmd("bg_as_blacklist_add", "concmd_AddToAsList", ACCESS_FLAG)
	register_concmd("bg_as_blacklist_del", "concmd_DelFromAsList", ACCESS_FLAG)
	register_concmd("bg_as_blacklist_show", "concmd_ShowAsList", ACCESS_FLAG)
	register_concmd("bg_as_whitelist_add", "concmd_AddToAsList", ACCESS_FLAG)
	register_concmd("bg_as_whitelist_del", "concmd_DelFromAsList", ACCESS_FLAG)
	register_concmd("bg_as_whitelist_show", "concmd_ShowAsList", ACCESS_FLAG)
	register_concmd("bg_ip_blacklist_add", "concmd_AddToIpList", ACCESS_FLAG)
	register_concmd("bg_ip_whitelist_add", "concmd_AddToIpList", ACCESS_FLAG)
	register_concmd("bg_ip_blacklist_del", "concmd_DelFromIpList", ACCESS_FLAG)
	register_concmd("bg_ip_whitelist_del", "concmd_DelFromIpList", ACCESS_FLAG)
	register_concmd("bg_ip_blacklist_show", "concmd_ShowIpList", ACCESS_FLAG)
	register_concmd("bg_ip_whitelist_show", "concmd_ShowIpList", ACCESS_FLAG)
	register_concmd("bg_find_ip", "concmd_FindIP", ACCESS_FLAG)
	register_concmd("bg_get_as_by_ip", "concmd_GetAsByIP", ACCESS_FLAG)
	register_concmd("bg_find_as", "concmd_FindAS", ACCESS_FLAG)
	register_concmd("bg_check_ip", "concmd_CheckIP", ACCESS_FLAG)
	register_concmd("bg_flush_data", "concmd_FlushData", ACCESS_FLAG)
	register_concmd("bg_status", "concmd_Status", ACCESS_FLAG)

	/* --- */

	new pCvar = get_cvar_pointer("amx_default_access")

	if(pCvar) {
		new szFlags[32]; get_pcvar_string(pCvar, szFlags, chx(szFlags))
		g_iDefAccFlags = read_flags(szFlags)
		hook_cvar_change(pCvar, "hook_CvarChange")
	}

	/* --- */

	g_eIpArray[LIST_TYPE__BLACKLIST] = ArrayCreate(RANGE_DATA_STRUCT, 1)
	g_eIpArray[LIST_TYPE__WHITELIST] = ArrayCreate(RANGE_DATA_STRUCT, 1)
	g_tAsCache = TrieCreate()
	g_tAsNumbers = TrieCreate()
	g_aAsArray[LIST_TYPE__BLACKLIST] = ArrayCreate(AS_DATA_STRUCT, 1)
	g_aAsArray[LIST_TYPE__WHITELIST] = ArrayCreate(AS_DATA_STRUCT, 1)
	g_tCurl = TrieCreate()
	g_tIpInfo = TrieCreate()
	g_tIpProxyStatus = TrieCreate()
	g_tIpMatch = TrieCreate()

	/* --- */

	new szLogDir[PLATFORM_MAX_PATH]

	get_localinfo("amxx_logs", szLogDir, chx(szLogDir))
	format(szLogDir, chx(szLogDir), "%s/%s", szLogDir, DIR_NAME)

	if(!dir_exists(szLogDir)) {
		mkdir(szLogDir)
	}

	for(new i; i <= LOG__LONG_QUERY; i++) {
		formatex(g_eLogFile[i], PLATFORM_MAX_PATH - 1, "%s/%s", szLogDir, LOG_NAME[i])
	}

	/* --- */

	new szCfgDir[PLATFORM_MAX_PATH]

	get_localinfo("amxx_configsdir", szCfgDir, chx(szCfgDir))
	format(szCfgDir, chx(szCfgDir), "%s/%s", szCfgDir, DIR_NAME)

	if(!dir_exists(szCfgDir)) {
		mkdir(szCfgDir)
	}

	new szKeyFile[PLATFORM_MAX_PATH]
	format(szKeyFile, chx(szKeyFile), "%s/%s", szCfgDir, KEY_FILE_NAME)
	func_LoadApiKeys(szKeyFile)

	format(g_szIpListFile, chx(g_szIpListFile), "%s/%s", szCfgDir, IP_FILE_NAME)
	format(g_szAsListFile, chx(g_szAsListFile), "%s/%s", szCfgDir, AS_FILE_NAME)
	func_LoadRanges()
	func_LoadAsNumbers()

	format(szCfgDir, chx(szCfgDir), "%s/%s", szCfgDir, COUNTRY_FILE_NAME)
	func_LoadAllowedCodes(szCfgDir)
}

/* -------------------- */

public OnConfigsExecuted() {
#if defined DEBUG
	log_message("[BG] Initializing...")
#endif

	/* --- */

	new szLogDir[PLATFORM_MAX_PATH]

	get_localinfo("amxx_logs", szLogDir, chx(szLogDir))
	format(szLogDir, chx(szLogDir), "%s/%s", szLogDir, DIR_NAME)

	new szTime[18], iDivide = get_pcvar_num(g_pCvar[PCVAR__DIVIDE_LOGS])

	if(iDivide) {
		get_time("%m.%Y_", szTime, chx(szTime))
	}

	for(new i = LOG__DENY; i <= LOG__ALLOW; i++) {
		formatex(g_eLogFile[i], PLATFORM_MAX_PATH - 1, "%s/%s%s", szLogDir, szTime, LOG_NAME[i])
	}

	/* --- */

	new szHost[SQL_ACC_LEN], szUser[SQL_ACC_LEN], szPwd[SQL_ACC_LEN], szDb[SQL_ACC_LEN]

	get_pcvar_string(g_pCvar[PCVAR__HOST], szHost, chx(szHost))
	get_pcvar_string(g_pCvar[PCVAR__USER], szUser, chx(szUser))
	get_pcvar_string(g_pCvar[PCVAR__PWD], szPwd, chx(szPwd))
	get_pcvar_string(g_pCvar[PCVAR__DB], szDb, chx(szDb))

#if defined AUTO_CFG
	new const szTemp[] = "*** PROTECTED ***"
	set_pcvar_string(g_pCvar[PCVAR__USER], szTemp)
	set_pcvar_string(g_pCvar[PCVAR__PWD], szTemp)
#endif

	get_pcvar_string(g_pCvar[PCVAR__SQL_MODE], g_szSqlMode, chx(g_szSqlMode))

	if(!SQL_SetAffinity(g_szSqlMode)) {
		plugin_end()
		set_fail_state("Failed to set affinity to '%s' (module not loaded?)", g_szSqlMode)
	}

	get_pcvar_string(g_pCvar[PCVAR__TABLE], g_szTableName, chx(g_szTableName))

	g_hSqlTuple = SQL_MakeDbTuple(szHost, szUser, szPwd, szDb, get_pcvar_num(g_pCvar[PCVAR__CONN_TIMEOUT]))

	SQL_SetCharset(g_hSqlTuple, "utf8")

	func_MakeInitQuery()
}

/* -------------------- */

stock func_MakeInitQuery() {
	set_task(g_eCvar[CVAR_F__REQUEST_TIMEOUT], "task_RetryInitQuery", TASKID__RETRY_INIT_QUERY)

	new iCreateTable = get_pcvar_num(g_pCvar[PCVAR__CREATE_TABLE])
	new	iPruneDays = get_pcvar_num(g_pCvar[PCVAR__PRUNE_OLD_IPS])
	new iLen

	if(equal(g_szSqlMode, "sqlite")) {
		if(iCreateTable) {
			iLen = formatex( g_szQuery, chx(g_szQuery),
				"CREATE TABLE IF NOT EXISTS `%s` (\
					`%s` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT UNIQUE, \
					`%s` INTEGER NOT NULL UNIQUE, \
					`%s` INTEGER NOT NULL, \
					`%s` TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP\
				);",

				g_szTableName,

				SQL_TABLE_COLUMNS[SQL_COLUMN__ID],
				SQL_TABLE_COLUMNS[SQL_COLUMN__IP],
				SQL_TABLE_COLUMNS[SQL_COLUMN__IP_STATUS],
				SQL_TABLE_COLUMNS[SQL_COLUMN__TIMESTAMP]
			);
		}

		if(iPruneDays > 0) {
			formatex( g_szQuery[iLen], chx(g_szQuery) - iLen,
				" DELETE FROM `%s` WHERE `%s` <= DATETIME('now','-%d day');",

				g_szTableName, SQL_TABLE_COLUMNS[SQL_COLUMN__TIMESTAMP], iPruneDays
			);
		}
	}
	else if(equal(g_szSqlMode, "mysql")) {
		if(iCreateTable) {
			iLen = formatex( g_szQuery, chx(g_szQuery),
				"CREATE TABLE IF NOT EXISTS `%s` (\
					`%s` int(10) unsigned NOT NULL AUTO_INCREMENT, \
					`%s` int(11) NOT NULL, \
					`%s` tinyint(1) unsigned NOT NULL, \
					`%s` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP, \
						PRIMARY KEY (`%s`), \
						UNIQUE KEY `%s` (`%s`)\
				) ENGINE=InnoDB DEFAULT CHARSET=utf8;",

				g_szTableName,

				SQL_TABLE_COLUMNS[SQL_COLUMN__ID],
				SQL_TABLE_COLUMNS[SQL_COLUMN__IP],
				SQL_TABLE_COLUMNS[SQL_COLUMN__IP_STATUS],
				SQL_TABLE_COLUMNS[SQL_COLUMN__TIMESTAMP],

				SQL_TABLE_COLUMNS[SQL_COLUMN__ID],
				SQL_TABLE_COLUMNS[SQL_COLUMN__IP], SQL_TABLE_COLUMNS[SQL_COLUMN__IP]
			);
		}

		if(iPruneDays > 0) {
			formatex( g_szQuery[iLen], chx(g_szQuery) - iLen,
				" DELETE FROM `%s` WHERE `%s` <= DATE_SUB(NOW(),INTERVAL %d DAY);",

				g_szTableName, SQL_TABLE_COLUMNS[SQL_COLUMN__TIMESTAMP], iPruneDays
			);
		}
	}
	else {
		plugin_end()
		set_fail_state("Invalid sql driver '%s' (valid are 'sqlite' or 'mysql')", g_szSqlMode)
	}

	func_MakeQuery(QUERY__INIT_SYSTEM)
}

/* -------------------- */

public task_RetryInitQuery() {
#if defined DEBUG
	log_message("[BG] Init query retrying...")
#endif
	static iCounter
	log_to_file(g_eLogFile[LOG__ERROR], "Init timeout #%d, retrying...", ++iCounter)
	func_MakeInitQuery()
}

/* -------------------- */

public SQL_Handler(iFailState, Handle:hQueryHandle, szError[], iErrorCode, eSqlData[], iDataSize, Float:fQueryTime) {
	// In order to avoid errors like ->
	// 		Run time error 10: native error (native "TrieGetString")
	// 		Invalid map handle provided (0)
	// This was found during testing. Apparently, the sql callback can be called after plugin_end()
	if(g_bPluginEnded) {
		return PLUGIN_HANDLED
	}

	if(iFailState != TQUERY_SUCCESS) {
		if(iFailState == TQUERY_CONNECT_FAILED)	{
		#if defined DEBUG
			log_message("[SQL] Can't connect to server [%.2f]", fQueryTime)
		#endif
			log_to_file(g_eLogFile[LOG__ERROR], "[SQL] Can't connect to server [%.2f]", fQueryTime)
			log_to_file(g_eLogFile[LOG__ERROR], "[SQL] Error #%d, %s", iErrorCode, szError)
		}
		else /*if(iFailState == TQUERY_QUERY_FAILED)*/ {
		#if defined DEBUG
			log_message("[SQL] Query error #%d, %s", iErrorCode, szError)
		#endif
			SQL_GetQueryString(hQueryHandle, g_szQuery, chx(g_szQuery))
			log_to_file(g_eLogFile[LOG__ERROR], "[SQL] Query error!")
			log_to_file(g_eLogFile[LOG__ERROR], "[SQL] Error #%d, %s", iErrorCode, szError)
			log_to_file(g_eLogFile[LOG__ERROR], "[SQL] Query: %s", g_szQuery)
		}

		if(g_eCvar[CVAR__KICK_IF_CANT_CHECK] && eSqlData[SQL_DATA__QUERY_TYPE] == QUERY__CHECK_IP) {
			new pPlayer = eSqlData[SQL_DATA__PLAYER_INDEX]

			TrieGetString(g_tIpMatch, fmt("%d", eSqlData[SQL_DATA__PLAYER_USERID]), g_szIP, chx(g_szIP))
			TrieDeleteKey(g_tIpMatch, fmt("%d", eSqlData[SQL_DATA__PLAYER_USERID]))

			if(!is_user_connected(pPlayer) || eSqlData[SQL_DATA__PLAYER_USERID] != get_user_userid(pPlayer)) {
				return PLUGIN_HANDLED
			}

			get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))
			func_KickPlayer(pPlayer, KICK_TYPE__PROXY_CHECK_FAIL)
		}

		return PLUGIN_HANDLED
	}

	if(fQueryTime > g_eCvar[CVAR_F__LONG_QUERY_TIME]) {
		SQL_GetQueryString(hQueryHandle, g_szQuery, chx(g_szQuery))
		log_to_file(g_eLogFile[LOG__LONG_QUERY], "[%.2f] %s", fQueryTime, g_szQuery)
	}

	/* --------- */

	switch(eSqlData[SQL_DATA__QUERY_TYPE]) {
		case QUERY__INIT_SYSTEM: {
			remove_task(TASKID__RETRY_INIT_QUERY)

			/*if(g_bSystemAvailable) {
				return PLUGIN_HANDLED
			}*/

			g_bSystemAvailable = true

		#if defined DEBUG
			log_message("[SQL] System initialized, gametime: %f", get_gametime())
		#endif
		}

		/* --- */

		case QUERY__CHECK_IP: {
			new pPlayer = eSqlData[SQL_DATA__PLAYER_INDEX]

			if(!is_user_connected(pPlayer) || eSqlData[SQL_DATA__PLAYER_USERID] != get_user_userid(pPlayer)) {
				pPlayer = 0
			}

			if(pPlayer) {
				get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))
			}

			if(!SQL_NumResults(hQueryHandle)) {
				func_CheckForProxy(pPlayer)
				return PLUGIN_HANDLED
			}

			TrieGetString(g_tIpMatch, fmt("%d", eSqlData[SQL_DATA__PLAYER_USERID]), g_szIP, chx(g_szIP))
			TrieDeleteKey(g_tIpMatch, fmt("%d", eSqlData[SQL_DATA__PLAYER_USERID]))

			new iIpStatus = SQL_ReadResult(hQueryHandle, 0)
			TrieSetCell(g_tIpProxyStatus, g_szIP, iIpStatus)

			if(pPlayer) {
				func_AgregateIpStatus(pPlayer, iIpStatus, AGREGATE_IP__SQL)
			}
		}

		/* --- */

		case QUERY__FLUSH_IPS: {
			TrieClear(g_tIpProxyStatus)
		}

		/* --- */

		// Next case here
	}

	return PLUGIN_HANDLED
}

/* -------------------- */

public client_putinserver(pPlayer) {
	g_szAccess[pPlayer] = _NA_
	g_szAsNumber[pPlayer] = _NA_
	g_szDesc[pPlayer] = _NA_

	if(!g_eCvar[CVAR__PLUGIN_ENABLED]) {
		g_szCode[pPlayer] = _NA_
		g_szCountry[pPlayer] = _NA_
		return
	}

	if(is_user_bot(pPlayer) || is_user_hltv(pPlayer)) {
		g_szAccess[pPlayer] = "Bot/HLTV"
		g_szCode[pPlayer] = _NA_
		g_szCountry[pPlayer] = _NA_
		return
	}

	get_user_ip(pPlayer, g_szIP, chx(g_szIP), .without_port = 1)

	func_GetCountryCode(pPlayer)
	func_GetCountryName(pPlayer)

	if(g_eCvar[CVAR__ALLOW_STEAM] && is_user_steam(pPlayer)) {
		g_szAccess[pPlayer] = "Steam"
		get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))

		log_to_file( g_eLogFile[LOG__ALLOW], "[Legit Steam] %n | %s | %s | %s | %s",
			pPlayer, g_szIP, g_szAuthID, g_szCode[pPlayer], g_szCountry[pPlayer] );

		return
	}

	set_task(Float:g_eCvar[CVAR__CHECK_DELAY], "task_CheckPlayer_Step1", pPlayer)
}

/* -------------------- */

public task_CheckPlayer_Step1(pPlayer) {
	if(!is_user_connected(pPlayer) || is_user_bot(pPlayer) || is_user_hltv(pPlayer)) {
		return
	}

	if(!g_bSystemAvailable) {
		set_task(2.0, "task_CheckPlayer_Step1", pPlayer)
		return
	}

	get_user_ip(pPlayer, g_szIP, chx(g_szIP), .without_port = 1)
	get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))

#if defined DEBUG
	new szIP[IP_LEN], iTimeStamp, iCurrIP = func_ParseIP(g_szIP)

	if(
		nvault_lookup(g_hDebug, fmt("%d", iCurrIP), szIP, sizeof(szIP), iTimeStamp)
			&&
		!equal(szIP, g_szIP)
	) {
		log_to_file(g_eLogFile[LOG__ERROR], "[DEBUG] Got int '%d' for '%s' and '%s'", iCurrIP, szIP, g_szIP)
	}
#endif

	new iFlags = get_user_flags(pPlayer)

	if(g_iDefAccFlags && !(iFlags & g_iDefAccFlags)) {
		g_szAccess[pPlayer] = "Access"

		new szFlags[32]; get_flags(iFlags, szFlags, chx(szFlags))

		log_to_file( g_eLogFile[LOG__ALLOW], "[Access Flags] %n | %s | %s | %s | %s | %s",
			pPlayer, g_szIP, g_szAuthID, szFlags, g_szCode[pPlayer], g_szCountry[pPlayer] );

		return
	}

	if(nvault_get(g_hImmunity, g_szAuthID)) {
		g_szAccess[pPlayer] = "Immunity"

		log_to_file( g_eLogFile[LOG__ALLOW], "[SteamID Immunity] %n | %s | %s | %s | %s",
			pPlayer, g_szIP, g_szAuthID, g_szCode[pPlayer], g_szCountry[pPlayer] );

		return
	}

	new iIP = func_ParseIP(g_szIP)

	if(IsIpInList(iIP, LIST_TYPE__WHITELIST)) {
		g_szAccess[pPlayer] = "IP Whitelist"
		new szStartIP[IP_LEN], szEndIP[IP_LEN]

		func_ReverseIP(g_eRangeData[RDS__START_IP], szStartIP, chx(szStartIP))
		func_ReverseIP(g_eRangeData[RDS__END_IP], szEndIP, chx(szEndIP))

		log_to_file( g_eLogFile[LOG__ALLOW], "[IP Whitelist] %n | %s | %s | %s - %s | %s | %s | %s",
			pPlayer, g_szIP, g_szAuthID, szStartIP, szEndIP, g_eRangeData[RDS__COMMENT], g_szCode[pPlayer], g_szCountry[pPlayer] );

		return
	}

	func_CheckAsNumber(pPlayer)
}

/* -------------------- */

stock func_CheckAsNumber(pPlayer) {
	g_eAsTrieData[AST__STATE] = STATE__NO
	TrieGetArray(g_tAsCache, g_szIP, g_eAsTrieData, sizeof(g_eAsTrieData))

	if(g_eAsTrieData[AST__STATE] == STATE__NO) {
		new eExtData[WHOIS_DATA_STRUCT]

		eExtData[WHOIS_DATA__QUERY_TYPE] = WHOIS_QUERY_TYPE__AUTO
		eExtData[WHOIS_DATA__QUERY_TRIES] = g_eCvar[CVAR__MAX_WHOIS_QUERY_TRIES]
		eExtData[WHOIS_DATA__PLAYER_USERID] = get_user_userid(pPlayer)

		WhoisIpQuery(g_szIP, "OnWhoisIpQueryComplete", eExtData, sizeof(eExtData))
		return
	}

	TrieGetArray(g_tIpInfo, g_szIP, g_eIpInfoData, sizeof(g_eIpInfoData))
	copy(g_szAsNumber[pPlayer], AS_LEN - 1, g_eIpInfoData[IP_INFO__AS])

	if(g_eIpInfoData[IP_INFO__DESC][0]) {
		copy(g_szDesc[pPlayer], DESC_LEN - 1, g_eIpInfoData[IP_INFO__DESC])
	}

	switch(g_eAsTrieData[AST__STATE]) {
		case STATE__NORMAL, STATE__WHITELIST: {
			func_CheckPlayer_Step2(pPlayer)
		}
		case STATE__BAN: {
			func_KickPlayer(pPlayer, KICK_TYPE__AS)
		}
	}
}

/* -------------------- */

public OnWhoisIpQueryComplete(WhoisQueryResult:iResult, szIP[], szStartIP[], szEndIP[], szAsNumber[], szDesc[], eExtData[], iExtDataSize/*, szRawAnswer[]*/) {
	// By analogy with SQL_Handler()
	if(g_bPluginEnded) {
		return
	}

	switch(eExtData[WHOIS_DATA__QUERY_TYPE]) {
		case WHOIS_QUERY_TYPE__AUTO: {
			new pPlayer = func_FindPlayer(eExtData[WHOIS_DATA__PLAYER_USERID], szIP)

			if(pPlayer) {
				// We need this data both for kick and for further check in func_CheckPlayer_Step2()
				copy(g_szIP, chx(g_szIP), szIP)
				get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))
			}

			g_eAsTrieData[AST__STATE] = STATE__NORMAL

			/* 	If 'szAsNumber' not present in 'g_tAsNumbers' (AS not banned), or it marked as "white", then in
				'g_tAsCache' will recorded 'g_eAsTrieData[AST__STATE]' with value 'STATE__NORMAL' or 'STATE__WHITELIST',
				which means value of 'g_eAsTrieData[AST__ARRAY_POS]' does not matter (will not be used).
				For the future: related logic below (STATE__BAN), and in 'func_CheckAsNumber()' */
			//g_eAsTrieData[AST__ARRAY_POS] = INVALID_HANDLE
			switch(iResult) {
				case WHOIS_NO_ERROR: {
					// 'g_eAsTrieData[AST__STATE]' does not change if 'szAsNumber' not present in 'g_tAsNumbers'
					TrieGetArray(g_tAsNumbers, szAsNumber, g_eAsTrieData, sizeof(g_eAsTrieData))

					TrieSetArray(g_tAsCache, szIP, g_eAsTrieData, sizeof(g_eAsTrieData))

					trim(szDesc)

					copy(g_eIpInfoData[IP_INFO__AS], AS_LEN - 1, szAsNumber)
					copy(g_eIpInfoData[IP_INFO__DESC], DESC_LEN - 1, szDesc)
					TrieSetArray(g_tIpInfo, szIP, g_eIpInfoData, sizeof(g_eIpInfoData))

					if(pPlayer) {
						copy(g_szAsNumber[pPlayer], AS_LEN - 1, szAsNumber)

						if(szDesc[0]) {
							copy(g_szDesc[pPlayer], DESC_LEN - 1, szDesc)
						}
					}

					if(g_eAsTrieData[AST__STATE] == STATE__BAN) {
						if(pPlayer) {
							func_KickPlayer(pPlayer, KICK_TYPE__AS)
						}

						return
					}
				}
				default: { // see whois.inc "WhoisQueryResult"
					if(--eExtData[WHOIS_DATA__QUERY_TRIES] == 0) {
						log_to_file( g_eLogFile[LOG__ERROR], "[Error #%d] Can't get WHOIS info for IP '%s' (%s, %s)",
							iResult, szIP, pPlayer ? g_szCode[pPlayer] : _NA_, pPlayer ? g_szCountry[pPlayer] : _NA_ );

						if(g_eCvar[CVAR__KICK_IF_CANT_CHECK] && pPlayer) {
							func_KickPlayer(pPlayer, KICK_TYPE__AS_CHECK_FAIL)
							return
						}
					}
					else {
						WhoisIpQuery(szIP, "OnWhoisIpQueryComplete", eExtData, iExtDataSize)
						return
					}
				}
			}

			if(pPlayer) {
				func_CheckPlayer_Step2(pPlayer)
			}
		}

		/* --- */

		case WHOIS_QUERY_TYPE__MANUAL: {
			copy(g_szGetWhoisIP, chx(g_szGetWhoisIP), szIP)

			switch(iResult) {
				case WHOIS_NO_ERROR: {
					trim(szDesc)
					copy(g_eIpInfoData[IP_INFO__AS], AS_LEN - 1, szAsNumber)
					copy(g_eIpInfoData[IP_INFO__DESC], DESC_LEN - 1, szDesc)
					TrieSetArray(g_tIpInfo, szIP, g_eIpInfoData, sizeof(g_eIpInfoData))
					g_iManualWhoisState = 1
				}
				default: { // see whois.inc "WhoisQueryResult"
					g_iManualWhoisState = INVALID_HANDLE
				}
			}
		}
	}
}

/* -------------------- */

stock func_CheckPlayer_Step2(pPlayer) {
	new iIP = func_ParseIP(g_szIP)

	if(IsIpInList(iIP, LIST_TYPE__BLACKLIST)) {
		func_KickPlayer(pPlayer, KICK_TYPE__IP_BAN)
		return
	}

	if(g_eAsTrieData[AST__STATE] == STATE__WHITELIST) {
		g_szAccess[pPlayer] = "AS Whitelist"

		ArrayGetArray(Array:g_aAsArray[LIST_TYPE__WHITELIST], g_eAsTrieData[AST__ARRAY_POS], g_eAsData)

		if(!g_eAsData[ASD__COMMENT][0]) {
			g_eAsData[ASD__COMMENT] = _NA_
		}

		log_to_file( g_eLogFile[LOG__ALLOW], "[AS Whitelist] %n | %s | %s | %s | %s | %s | %s | %s",
			pPlayer, g_szIP, g_szAuthID, g_szAsNumber[pPlayer], g_szDesc[pPlayer],
			g_eAsData[ASD__COMMENT], g_szCode[pPlayer], g_szCountry[pPlayer] );

		return
	}

	if(equal(g_szCode[pPlayer], _NA_)) {
		func_KickPlayer(pPlayer, KICK_TYPE__COUNTRY)
		return
	}

	for(new i; i < g_iCodesCount; i++) {
		if(!equal(g_szCode[pPlayer], g_szAllowedCodes[i])) {
			continue
		}

		new iIpStatus

		if(TrieGetCell(g_tIpProxyStatus, g_szIP, iIpStatus)) {
			func_AgregateIpStatus(pPlayer, iIpStatus, AGREGATE_IP__CACHE)
			return
		}

		formatex( g_szQuery, chx(g_szQuery), "SELECT `%s` FROM `%s` WHERE `%s` = '%d' LIMIT 1",
			SQL_TABLE_COLUMNS[SQL_COLUMN__IP_STATUS], g_szTableName, SQL_TABLE_COLUMNS[SQL_COLUMN__IP], func_ParseIP(g_szIP) );

		g_eSqlData[SQL_DATA__PLAYER_INDEX] = pPlayer
		g_eSqlData[SQL_DATA__PLAYER_USERID] = get_user_userid(pPlayer)
		TrieSetString(g_tIpMatch, fmt("%d", g_eSqlData[SQL_DATA__PLAYER_USERID]), g_szIP)

		func_MakeQuery(QUERY__CHECK_IP)
		return
	}

	func_KickPlayer(pPlayer, KICK_TYPE__COUNTRY)
}

/* -------------------- */

stock func_AgregateIpStatus(pPlayer, iIpStatus, iAgregateType) {
	if(iIpStatus == STATE__NORMAL || (!g_eCvar[CVAR__BAN_SUSPICIOUS] && iIpStatus == STATE__SUSPICIOUS)) {
		static const AGREGATE_TYPE[AGREGATE_TYPE_ENUM][] = { "Cache:", "SQL:" }
		copy(g_szAccess[pPlayer], chx(g_szAccess[]), AGREGATE_TYPE[iAgregateType])
		add(g_szAccess[pPlayer], chx(g_szAccess[]), (iIpStatus == STATE__NORMAL) ? " Normal" : " Susp.")
		func_LogAllow(pPlayer)

		return
	}

	// if(iIpStatus == STATE__BAN || iIpStatus == STATE__SUSPICIOUS) ->

	static const KICK_TYPE_ENUM:KICK_TYPE[AGREGATE_TYPE_ENUM][2] = {
		{ KICK_TYPE__PROXY_CACHE, KICK_TYPE__SUSPICIOUS_CACHE },
		{ KICK_TYPE__PROXY_SQL, KICK_TYPE__SUSPICIOUS_SQL }
	}

	func_KickPlayer(pPlayer, KICK_TYPE[iAgregateType][ (iIpStatus == STATE__BAN) ? 0 : 1 ])
}

/* -------------------- */

stock func_CheckForProxy(pPlayer) {
	new iUserID = (pPlayer > 0) ? get_user_userid(pPlayer) : random_num(1000, 2000)

	formatex(g_szFile, chx(g_szFile), "%s/%s_%d", g_szDataDir, g_szIP, iUserID)

	/* The file may exist if the request is sent, the file was created for
		writing, and the map was changed before the end of the check */
	if(file_exists(g_szFile)) {
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] File '%s' already exists!", g_szFile)
		delete_file(g_szFile)

		if(g_eCvar[CVAR__KICK_IF_CANT_CHECK] && pPlayer > 0) { // pPlayer can be -1, 0, 1-MaxClients
			func_KickPlayer(pPlayer, KICK_TYPE__PROXY_CHECK_FAIL)
		}

		return
	}

	new eData[CURL_DATA_STRUCT]
	eData[CURL_DATA__FILE_HANDLE] = fopen(g_szFile, "wb")

	if(!eData[CURL_DATA__FILE_HANDLE]) {
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open '%s' for writing!", g_szFile)
		return
	}

	eData[CURL_DATA__PLAYER_INDEX] = pPlayer
	eData[CURL_DATA__PLAYER_USERID] = iUserID

	new CURL:hCurl = curl_easy_init()

	curl_easy_setopt(hCurl, CURLOPT_BUFFERSIZE, CURL_BUFFER_SIZE)
	formatex(g_szRequest, chx(g_szRequest), "http://v2.api.iphub.info/ip/%s", g_szIP)
	curl_easy_setopt(hCurl, CURLOPT_URL, g_szRequest)

	new curl_slist:hList
	hList = curl_slist_append(hList, g_szApiKey[g_iCurrentKey])
	curl_easy_setopt(hCurl, CURLOPT_HTTPHEADER, hList)

	curl_easy_setopt(hCurl, CURLOPT_WRITEDATA, eData[CURL_DATA__FILE_HANDLE])
	curl_easy_setopt(hCurl, CURLOPT_WRITEFUNCTION, "OnWrite")

	eData[CURL_DATA__SLIST_HANDLE] = hList
	TrieSetString(g_tCurl, fmt("%d", hCurl), g_szIP)
	curl_easy_perform(hCurl, "OnDownloadComplete", eData, sizeof(eData))
}

/* -------------------- */

public OnWrite(szData[], iSize, nmemb, hFile) {
	new iActualSize = iSize * nmemb

	fwrite_blocks(hFile, szData, iActualSize, BLOCK_CHAR)

	return iActualSize
}

/* -------------------- */

public OnDownloadComplete(CURL:hCurl, CURLcode:iCode, eData[]/*, iDataSize*/) { // iDataSize always = 0 (nothing passed?)
	/* "Also, if there are active transfers at the time of the map change, the module will try
	generate a transmission error to complete the stream as soon as possible, and call the
	curl_easy_perform callback (called after plugin_end())"
	http://amx-x.ru/viewtopic.php?f=21&t=31672 */
	if(g_bPluginEnded) {
		return PLUGIN_HANDLED
	}

	TrieGetString(g_tCurl, fmt("%d", hCurl), g_szIP, chx(g_szIP))
	TrieDeleteKey(g_tCurl, fmt("%d", hCurl))

	new iUserID = eData[CURL_DATA__PLAYER_USERID]

	formatex(g_szFile, chx(g_szFile), "%s/%s_%d", g_szDataDir, g_szIP, iUserID)

	fclose(eData[CURL_DATA__FILE_HANDLE])
	curl_slist_free_all(curl_slist:eData[CURL_DATA__SLIST_HANDLE])

	curl_easy_cleanup(hCurl)

	new pPlayerIndex = eData[CURL_DATA__PLAYER_INDEX]

	if(iCode != CURLE_OK) {
		new szError[256]; curl_easy_strerror(iCode, szError, chx(szError))
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] Curl error code #%d, IP '%s'", iCode, g_szIP)
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] Curl error: %s", szError)
		return func_CurlError(pPlayerIndex, iUserID)
	}

	/* --- */

	if(!file_exists(g_szFile)) {
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] File '%s' not found!", g_szFile)
		return func_CurlError(pPlayerIndex, iUserID)
	}

	new JSON:hJson = json_parse(g_szFile, true, true)

	if(hJson == Invalid_JSON) {
		new bool:bError, szBuffer[256], iFile = fopen(g_szFile, "r")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open wrong file '%s' for analysis", g_szFile)
		}
		else {
			while(!feof(iFile)) {
				fgets(iFile, szBuffer, chx(szBuffer))

				/*  <html>
					<head><title>502 Bad Gateway</title></head>
					<body>
					<center><h1>502 Bad Gateway</h1></center>
					<hr><center>nginx/1.15.8</center>
					</body>
					</html> */
				if(contain(szBuffer, "502") != -1) {
					bError = true
					break
				}
			}

			fclose(iFile)

			if(bError) {
				log_to_file(g_eLogFile[LOG__ERROR], "[Error] HTTP responce code 502 (Bad Gateway)", g_szFile)
			}
			else {
				log_to_file(g_eLogFile[LOG__ERROR], "[Error] Something wrong with file '%s'", g_szFile)
			}
		}

		return func_CurlError(pPlayerIndex, iUserID)
	}

	new iCount = json_object_get_count(hJson)

	if(iCount < EXPECTED_JSON_KEY_COUNT) {

		json_free(hJson)

		new iError, szBuffer[256], iFile = fopen(g_szFile, "r")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open wrong file '%s' for analysis", g_szFile)
		}
		else {
			while(!feof(iFile)) {
				fgets(iFile, szBuffer, chx(szBuffer))

				// {"code":"TooManyRequests","message":"You've exceeded your rate limit of 1000 request(s) per 86400 second(s)."}
				if(contain(szBuffer, "TooManyRequests") != -1) {
					iError = 1
					break
				}

				// {"code":"Forbidden","message":"Invalid API key"}
				// {"code":"Forbidden","message":"Missing X-Key header"}
				if(contain(szBuffer, "Forbidden") != -1) {
					iError = 2
					break
				}
			}

			fclose(iFile)

			switch(iError) {
				case 0: {
					log_to_file(g_eLogFile[LOG__ERROR], "[Error] Wrong key count (%d/%d) in file '%s'", iCount, EXPECTED_JSON_KEY_COUNT, g_szFile)
				}
				case 1: {
					if(g_iCurrentKey < g_iLoadedKeys - 1) {
						log_to_file( g_eLogFile[LOG__ERROR], "[Error] Request limit reached for API key '%d', switching to API key '%d' of '%d'",
							g_iCurrentKey + 1, ++g_iCurrentKey, g_iLoadedKeys );
					}
					else {
						g_iCurrentKey = 0
						log_to_file(g_eLogFile[LOG__ERROR], "[Error] Request limit reached, and no more available API keys!")
					}
				}
				case 2: {
					log_to_file(g_eLogFile[LOG__ERROR], "[Error] Invalid API Key!")
				}
			}
		}

		return func_CurlError(pPlayerIndex, iUserID)
	}

	new szKey[32]

	json_object_get_name(hJson, JSON_KEY__CHECK_STATE, szKey, chx(szKey))

	new JSON:hVal = json_object_get_value(hJson, szKey, false)

	if(hVal == Invalid_JSON) {
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] Json object 'JSON_KEY__CHECK_STATE' is invalid (IP %s)", g_szIP)
		json_free(hJson)
		return func_CurlError(pPlayerIndex, iUserID)
	}

	if(json_get_type(hVal) != JSONNumber) {
		log_to_file(g_eLogFile[LOG__ERROR], "[Error] Json object 'JSON_KEY__CHECK_STATE' is not a number (IP %s)", g_szIP)
		json_free(hVal)
		json_free(hJson)
		return func_CurlError(pPlayerIndex, iUserID)
	}

	new iState = json_get_number(hVal)

	json_free(hVal)
	json_free(hJson)

	/* --- */

	new pPlayer = func_FindPlayer(iUserID, g_szIP)

	if(!pPlayer) {
		func_GetCountryCode(pPlayer)
		func_GetCountryName(pPlayer)

		if(TrieGetArray(g_tIpInfo, g_szIP, g_eIpInfoData, sizeof(g_eIpInfoData))) {
			copy(g_szAsNumber[0], AS_LEN - 1, g_eIpInfoData[IP_INFO__AS])

			if(g_eIpInfoData[IP_INFO__DESC][0]) {
				copy(g_szDesc[0], DESC_LEN - 1, g_eIpInfoData[IP_INFO__DESC])
			}
		}
		else {
			g_szAsNumber[0] = _NA_
			g_szDesc[0] = _NA_
		}
	}

	if(iState == IP_CHECK_STATE__NORMAL || (!g_eCvar[CVAR__BAN_SUSPICIOUS] && iState == IP_CHECK_STATE__SUSPICIOUS)) {
		new iIpState = (iState == IP_CHECK_STATE__NORMAL) ? STATE__NORMAL : STATE__SUSPICIOUS;
		new bool:bKeySet = bool:TrieSetCell(g_tIpProxyStatus, g_szIP, iIpState, .replace = false)

		if(pPlayerIndex == INVALID_HANDLE) {
			copy(g_szGetProxyIP, chx(g_szGetProxyIP), g_szIP)
		}
		else {
			if(bKeySet) { // Prevent sql insert error, that may occur if two players with same IP connects ~ at the same time
				func_WriteIpToSqlTable(iIpState)
			}

			log_to_file( g_eLogFile[LOG__PROXY_CHECK], "[Allow: %s] %s | %s | %s | %s | %s | %s",
				iState == IP_CHECK_STATE__NORMAL ? "Normal" : "Susp.", func_GetPlayerName(pPlayer), g_szIP,
				g_szCode[pPlayer], g_szCountry[pPlayer], g_szAsNumber[pPlayer], g_szDesc[pPlayer] );

			if(pPlayer) {
				if(iState == IP_CHECK_STATE__NORMAL) {
					g_szAccess[pPlayer] = "Check: Normal"
				}
				else {
					g_szAccess[pPlayer] = "Check: Susp."
				}

				func_LogAllow(pPlayer)
			}
		}

		delete_file(g_szFile)
		return PLUGIN_HANDLED
	}

	if(iState == IP_CHECK_STATE__DETECT || iState == IP_CHECK_STATE__SUSPICIOUS) {
		new iIpState = (iState == IP_CHECK_STATE__DETECT) ? STATE__BAN : STATE__SUSPICIOUS;
		new bool:bKeySet = bool:TrieSetCell(g_tIpProxyStatus, g_szIP, iIpState, .replace = false)

		if(pPlayerIndex == INVALID_HANDLE) {
			copy(g_szGetProxyIP, chx(g_szGetProxyIP), g_szIP)
		}
		else {
			if(bKeySet) { // Prevent sql insert error, that may occur if two players with same IP connects ~ at the same time
				func_WriteIpToSqlTable(STATE__BAN)
			}

			log_to_file( g_eLogFile[LOG__PROXY_CHECK], "[%s] %s | %s | %s | %s | %s | %s",
				iState == IP_CHECK_STATE__DETECT ? "Proxy/VPN" : "Deny: Susp.", func_GetPlayerName(pPlayer),
				g_szIP,	g_szCode[pPlayer], g_szCountry[pPlayer], g_szAsNumber[pPlayer], g_szDesc[pPlayer] );

			if(pPlayer) {
				get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))
				func_KickPlayer(pPlayer, (iState == IP_CHECK_STATE__DETECT) ? KICK_TYPE__PROXY_CHECK : KICK_TYPE__SUSPICIOUS_CHECK)
			}
		}

		delete_file(g_szFile)
		return PLUGIN_HANDLED
	}

	log_to_file( g_eLogFile[LOG__ERROR], "[Error] Wrong check state value '%d', IP '%s' (%s | %s | %s | %s)",
		iState, g_szIP, g_szCode[pPlayer], g_szCountry[pPlayer], g_szAsNumber[pPlayer], g_szDesc[pPlayer] );

	return func_CurlError(pPlayerIndex, iUserID)
}

/* -------------------- */

stock func_WriteIpToSqlTable(iIpState) {
#if defined DEBUG
	new iIP = func_ParseIP(g_szIP)
	nvault_set(g_hDebug, fmt("%d", iIP), g_szIP)
#endif

	formatex( g_szQuery, chx(g_szQuery), "INSERT INTO `%s` (`%s`, `%s`) VALUES ('%d', '%d')",
		g_szTableName, SQL_TABLE_COLUMNS[SQL_COLUMN__IP], SQL_TABLE_COLUMNS[SQL_COLUMN__IP_STATUS],
		func_ParseIP(g_szIP), iIpState
	);

	func_MakeQuery(QUERY__INSERT_IP)
}

/* -------------------- */

stock func_CurlError(pPlayerIndex, iUserID) {
	new szNewFileName[PLATFORM_MAX_PATH]
	formatex(szNewFileName, chx(szNewFileName), "%s_error", g_szFile)
	rename_file(g_szFile, szNewFileName, .relative = 1)

	if(!g_eCvar[CVAR__KICK_IF_CANT_CHECK] || pPlayerIndex == INVALID_HANDLE) {
		return PLUGIN_HANDLED
	}

	new pPlayer = func_FindPlayer(iUserID, g_szIP)

	if(pPlayer) {
		get_user_authid(pPlayer, g_szAuthID, chx(g_szAuthID))
		func_KickPlayer(pPlayer, KICK_TYPE__PROXY_CHECK_FAIL)
	}

	return PLUGIN_HANDLED
}

/* -------------------- */

stock func_FindPlayer(iUserID, szIP[]) {
	new pPlayer = find_player("k", iUserID)

	if(!pPlayer) {
		pPlayer = find_player("d", szIP)
	}

	return pPlayer
}

/* -------------------- */

stock func_KickPlayer(pPlayer, KICK_TYPE_ENUM:iKickType) {
	static const KICK_TYPE[KICK_TYPE_ENUM][] = {
		"Banned AS",
		"IP Blacklist",
		"Bad Country",
		"Check: Proxy/VPN",
		"SQL: Proxy/VPN",
		"Cache: Proxy/VPN",
		"Check: Suspicious IP",
		"SQL: Suspicious IP",
		"Cache: Suspicious IP",
		"AS Check Fail",
		"Proxy Check Fail"
	}

	new szMsg[320]

	new iLen = formatex(szMsg, chx(szMsg), "[%s] %n | %s | %s", KICK_TYPE[iKickType], pPlayer, g_szIP, g_szAuthID)

	switch(iKickType) {
		case KICK_TYPE__AS: {
			ArrayGetArray(Array:g_aAsArray[LIST_TYPE__BLACKLIST], g_eAsTrieData[AST__ARRAY_POS], g_eAsData)

			if(!g_eAsData[ASD__COMMENT][0]) {
				g_eAsData[ASD__COMMENT] = _NA_
			}

			iLen += formatex(szMsg[iLen], chx(szMsg) - iLen, " | %s | %s | %s", g_szAsNumber[pPlayer], g_szDesc[pPlayer], g_eAsData[ASD__COMMENT])
		}
		case KICK_TYPE__IP_BAN: {
			new szStartIP[IP_LEN], szEndIP[IP_LEN]

			func_ReverseIP(g_eRangeData[RDS__START_IP], szStartIP, chx(szStartIP))
			func_ReverseIP(g_eRangeData[RDS__END_IP], szEndIP, chx(szEndIP))

			if(!g_eRangeData[RDS__COMMENT][0]) {
				g_eRangeData[RDS__COMMENT] = _NA_
			}

			iLen += formatex( szMsg[iLen], chx(szMsg) - iLen, " | %s - %s | %s | %s | %s",
				szStartIP, szEndIP, g_eRangeData[RDS__COMMENT], g_szAsNumber[pPlayer], g_szDesc[pPlayer] );
		}
		case KICK_TYPE__AS_CHECK_FAIL: {
			// AS and Desc not obtained, no reason to print "N/A"
		}
		default: { // KICK_TYPE__COUNTRY, KICK_TYPE__SUSPICIOUS_..., KICK_TYPE__PROXY_..., KICK_TYPE__PROXY_CHECK_FAIL
			iLen += formatex(szMsg[iLen], chx(szMsg) - iLen, " | %s | %s", g_szAsNumber[pPlayer], g_szDesc[pPlayer])
		}
	}

	formatex(szMsg[iLen], chx(szMsg) - iLen, " | %s | %s", g_szCode[pPlayer], g_szCountry[pPlayer])

	log_to_file(g_eLogFile[LOG__DENY], szMsg)

	console_print(pPlayer, "^n%L %n | %s | %s", pPlayer, "BG__YOUR_NAME_STEAMID_AND_IP", pPlayer, g_szAuthID, g_szIP)

	if(g_eCvar[CVAR__SHOW_URL]) {
		console_print(pPlayer, "%L^n", pPlayer, "BG__URL")
	}

	new iData[1]

	switch(iKickType) {
		case KICK_TYPE__AS_CHECK_FAIL, KICK_TYPE__PROXY_CHECK_FAIL: {
			iData[0] = 1
		}
	}

	set_task(Float:g_eCvar[CVAR__KICK_DELAY], "task_DelayedKick", pPlayer, iData, sizeof(iData))
}

/* -------------------- */

public task_DelayedKick(iData[], pPlayer) {
	if(!is_user_connected(pPlayer)) {
		return
	}

	server_cmd("kick #%d ^"%L^"", get_user_userid(pPlayer), pPlayer, iData[0] ? "BG__KICK_REASON_CHECK_FAILED" : "BG__KICK_REASON_ACCESS")
}

/* -------------------- */

stock func_GetPlayerName(pPlayer) {
	new szName[MAX_NAME_LENGTH]

	if(pPlayer) {
		get_user_name(pPlayer, szName, chx(szName))
	}
	else {
		copy(szName, chx(szName), _NA_)
	}

	return szName
}

/* -------------------- */

public concmd_AllowSteamID(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(read_argc() == 1) {
		console_print(pPlayer, "* Usage: bg_allow_steamid <steamid> - Add immunity to all checks by SteamID")
		return PLUGIN_HANDLED
	}

	new szBuffer[64]

	read_args(szBuffer, chx(szBuffer))
	remove_quotes(szBuffer)
	trim(szBuffer)

	if(strlen(szBuffer) > AUTHID_LEN || !IsValidSteamID(szBuffer)) {
		console_print(pPlayer, "* Wrong SteamID provided, check your input!")
	}
	else if(nvault_get(g_hImmunity, szBuffer)) {
		console_print(pPlayer, "* SteamID '%s' already exists in immunity list!", szBuffer)
	}
	else {
		nvault_set(g_hImmunity, szBuffer, "1")
		console_print(pPlayer, "* SteamID '%s' added to immunity list", szBuffer)
		log_to_file(g_eLogFile[LOG__CMD], "Steamid '%s' added to immunity list by %N", szBuffer, pPlayer)
	}

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_AddToAsList(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	new szBuffer[64], iType

	read_argv(0, szBuffer, chx(szBuffer))

	if(equali(szBuffer, "bg_as_blacklist_add")) {
		iType = LIST_TYPE__BLACKLIST
	}
	else {
		iType = LIST_TYPE__WHITELIST
	}

	new iArgCount = read_argc()

	if(iArgCount == 1) {
		console_print( pPlayer, "* Usage: %s <as_number> <^"comment^">[optional]",
			iType == LIST_TYPE__BLACKLIST ? "bg_as_blacklist_add" : "bg_as_whitelist_add" );

		return PLUGIN_HANDLED
	}

	read_argv(1, szBuffer, chx(szBuffer))

	if(!IsValidAsNumber(szBuffer)) {
		console_print(pPlayer, "* Wrong AS number, check your input!")
		return PLUGIN_HANDLED
	}

	g_eAsTrieData[AST__STATE] = STATE__NO
	TrieGetArray(g_tAsNumbers, szBuffer, g_eAsTrieData, sizeof(g_eAsTrieData))

	if(g_eAsTrieData[AST__STATE] != STATE__NO) {
		console_print( pPlayer, "* AS number '%s' already in %s!", szBuffer,
			LIST_NAME[g_eAsTrieData[AST__STATE] == STATE__BAN ? LIST_TYPE__BLACKLIST : LIST_TYPE__WHITELIST] );

		return PLUGIN_HANDLED
	}

	copy(g_eAsData[ASD__NUMBER], AS_LEN - 1, szBuffer)

	if(iArgCount > 2) {
		read_argv(2, szBuffer, chx(szBuffer))
		trim(szBuffer)

		if(strlen(szBuffer) > COMMENT_LEN) {
			console_print(pPlayer, "* Comment is too long, max %d chars!", COMMENT_LEN)
			return PLUGIN_HANDLED
		}
	}
	else {
		szBuffer[0] = EOS
	}

	copy(g_eAsData[ASD__COMMENT], COMMENT_LEN - 1, szBuffer)

	new iFile = fopen(g_szAsListFile, "a")

	if(!iFile) {
		console_print(pPlayer, "* Error, can't open AS list file!")
		return PLUGIN_HANDLED
	}

	fprintf(iFile, "^n%s %s", LIST_NAME[iType], g_eAsData[ASD__NUMBER])

	if(szBuffer[0]) {
		fprintf(iFile, " ^"%s^"", szBuffer)
	}

	fclose(iFile)

	TrieClear(g_tAsCache)

	g_eAsTrieData[AST__STATE] = (iType == LIST_TYPE__BLACKLIST) ? STATE__BAN : STATE__WHITELIST;
	g_eAsTrieData[AST__ARRAY_POS] = g_iAsCount[iType]
	TrieSetArray(g_tAsNumbers, g_eAsData[ASD__NUMBER], g_eAsTrieData, sizeof(g_eAsTrieData))

	ArrayPushArray(g_aAsArray[iType], g_eAsData)
	g_iAsCount[iType]++

	console_print(pPlayer, "* AS number '%s' successfully added to %s!", g_eAsData[ASD__NUMBER], LIST_NAME[iType])

	log_to_file(g_eLogFile[LOG__CMD], "AS number '%s' added to %s by %N", g_eAsData[ASD__NUMBER], LIST_NAME[iType], pPlayer)

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_DelFromAsList(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	new szBuffer[64], iType

	read_argv(0, szBuffer, chx(szBuffer))

	if(equali(szBuffer, "bg_as_blacklist_del")) {
		iType = LIST_TYPE__BLACKLIST
	}
	else {
		iType = LIST_TYPE__WHITELIST
	}

	if(read_argc() == 1) {
		console_print( pPlayer, "* Usage: %s <as_number>",
			iType == LIST_TYPE__BLACKLIST ? "bg_as_blacklist_del" : "bg_as_whitelist_del" );

		return PLUGIN_HANDLED
	}

	read_argv(1, szBuffer, chx(szBuffer))

	if(!IsValidAsNumber(szBuffer)) {
		console_print(pPlayer, "* Wrong AS number, check your input!")
		return PLUGIN_HANDLED
	}

	g_eAsTrieData[AST__STATE] = STATE__NO
	TrieGetArray(g_tAsNumbers, szBuffer, g_eAsTrieData, sizeof(g_eAsTrieData))

	if(g_eAsTrieData[AST__STATE] == STATE__NO) {
		console_print(pPlayer, "* AS number '%s' NOT found in %s!", szBuffer, LIST_NAME[iType])
		return PLUGIN_HANDLED
	}

	TrieClear(g_tAsCache)
	TrieClear(g_tAsNumbers)
	g_iAsCount[LIST_TYPE__BLACKLIST] = 0
	g_iAsCount[LIST_TYPE__WHITELIST] = 0
	ArrayClear(Array:g_aAsArray[LIST_TYPE__BLACKLIST])
	ArrayClear(Array:g_aAsArray[LIST_TYPE__WHITELIST])

	new szOldFileName[PLATFORM_MAX_PATH]
	formatex(szOldFileName, chx(szOldFileName), "%s_old", g_szAsListFile)

	if(!rename_file(g_szAsListFile, szOldFileName, .relative = 1)) {
		console_print(pPlayer, "* Error, can't rename AS list file!")
		return PLUGIN_HANDLED
	}

	new iOldFile = fopen(szOldFileName, "r")

	if(!iOldFile) {
		console_print(pPlayer, "* Error, can't open renamed AS list file!")
		return PLUGIN_HANDLED
	}

	new iNewFile = fopen(g_szAsListFile, "w")

	if(!iNewFile) {
		fclose(iOldFile)
		console_print(pPlayer, "* Error, can't open new AS list file!")
		return PLUGIN_HANDLED
	}

	new szString[128], bool:bFound, iCfgType

	func_AddDefSting_AS(iNewFile)

	while(!feof(iOldFile)) {
		fgets(iOldFile, szString, chx(szString))
		trim(szString)

		switch(szString[0]) {
			case 'b': iCfgType = LIST_TYPE__BLACKLIST
			case 'w': iCfgType = LIST_TYPE__WHITELIST
			default: continue //  ';', '/', EOS, etc.
		}

		g_eAsData[ASD__COMMENT][0] = EOS

		parse(szString, "", "", g_eAsData[ASD__NUMBER], AS_LEN - 1, g_eAsData[ASD__COMMENT], COMMENT_LEN - 1)

		if(equal(g_eAsData[ASD__NUMBER], szBuffer)) {
			bFound = true
			continue
		}

		fprintf(iNewFile, "^n%s %s", LIST_NAME[iCfgType], g_eAsData[ASD__NUMBER])

		if(g_eAsData[ASD__COMMENT][0]) {
			fprintf(iNewFile, " ^"%s^"", g_eAsData[ASD__COMMENT])
		}

		g_eAsTrieData[AST__STATE] = (iCfgType == LIST_TYPE__BLACKLIST) ? STATE__BAN : STATE__WHITELIST;
		g_eAsTrieData[AST__ARRAY_POS] = g_iAsCount[iCfgType]
		TrieSetArray(g_tAsNumbers, g_eAsData[ASD__NUMBER], g_eAsTrieData, sizeof(g_eAsTrieData))

		ArrayPushArray(g_aAsArray[iCfgType], g_eAsData)
		g_iAsCount[iCfgType]++
	}

	fclose(iOldFile)
	fclose(iNewFile)

	if(!delete_file(szOldFileName)) {
		console_print(pPlayer, "* Warning! Can't delete '%s'", szOldFileName)
	}

	if(bFound) {
		console_print(pPlayer, "* AS number '%s' successfully removed from %s!", szBuffer, LIST_NAME[iType])
	}
	else {
		console_print(pPlayer, "* AS number '%s' removed from memory, but not found in '%s'!", szBuffer, AS_FILE_NAME)
	}

	log_to_file(g_eLogFile[LOG__CMD], "AS number '%s' removed from %s by %N", szBuffer, LIST_NAME[iType], pPlayer)

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_ShowAsList(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	new szBuffer[64], iType

	read_argv(0, szBuffer, chx(szBuffer))

	if(equali(szBuffer, "bg_as_blacklist_show")) {
		iType = LIST_TYPE__BLACKLIST
	}
	else {
		iType = LIST_TYPE__WHITELIST
	}

	new iTotalPages = floatround(float(g_iAsCount[iType]) / 10.0, floatround_ceil)

	if(read_argc() == 1) {
		console_print(pPlayer, "* Usage: %s <page> (total pages: %d)",
			iType == LIST_TYPE__BLACKLIST ? "bg_as_blacklist_show" : "bg_as_whitelist_show", iTotalPages );

		return PLUGIN_HANDLED
	}

	new iPage = read_argv_int(1)

	if(iPage < 1) {
		iPage = 1
	}

	if(iPage > iTotalPages) {
		iPage = iTotalPages
	}

	if(!iPage) {
		console_print(pPlayer, "* AS %s is empty!", LIST_NAME[iType])
		return PLUGIN_HANDLED
	}

	console_print(pPlayer, "Displaying page %d/%d:", iPage, iTotalPages)
	console_print(pPlayer, "# <-> AS number <-> Comment")

	new iCount, iStartPos, i
	iStartPos = i = 10 * (iPage - 1)

	while(i < g_iAsCount[iType]) {
		ArrayGetArray(g_aAsArray[iType], i, g_eAsData)

		if(!g_eAsData[ASD__COMMENT][0]) {
			g_eAsData[ASD__COMMENT] = _NA_
		}

		console_print(pPlayer, "%d <-> %s <-> %s", i + 1, g_eAsData[ASD__NUMBER], g_eAsData[ASD__COMMENT])

		if(++iCount == 10) {
			break
		}

		i++
	}

	console_print(pPlayer, "Displayed records: %d/%d", iStartPos + iCount, g_iAsCount[iType])

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_AddToIpList(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	new szBuffer[64], iType

	read_argv(0, szBuffer, chx(szBuffer))

	if(equali(szBuffer, "bg_ip_blacklist_add")) {
		iType = LIST_TYPE__BLACKLIST
	}
	else {
		iType = LIST_TYPE__WHITELIST
	}

	new iArgCount = read_argc()

	if(iArgCount < 3) {
		console_print( pPlayer, "* Usage: %s <start_ip> <end_ip> <^"comment^">[optional]",
			iType == LIST_TYPE__BLACKLIST ? "bg_ip_blacklist_add" : "bg_ip_whitelist_add" );

		return PLUGIN_HANDLED
	}

	new szStart[IP_LEN], szEnd[IP_LEN]

	read_argv(1, szStart, chx(szStart))
	read_argv(2, szEnd, chx(szEnd))

	if(!IsIpValid(szStart) || !IsIpValid(szEnd)) {
		console_print(pPlayer, "* Wrong value, check your input!")
		return PLUGIN_HANDLED
	}

	g_eRangeData[RDS__START_IP] = func_ParseIP(szStart)
	g_eRangeData[RDS__END_IP] = func_ParseIP(szEnd)

	if((g_eRangeData[RDS__END_IP] - g_eRangeData[RDS__START_IP] + 1) < 1) {
		console_print(pPlayer, "* Start IP is lower than End IP!")
		return PLUGIN_HANDLED
	}

	if(iArgCount > 3) {
		read_argv(3, szBuffer, chx(szBuffer))
		trim(szBuffer)

		if(strlen(szBuffer) > COMMENT_LEN) {
			console_print(pPlayer, "* Comment is too long, max %d chars!", COMMENT_LEN)
			return PLUGIN_HANDLED
		}
	}
	else {
		szBuffer[0] = EOS
	}

	copy(g_eRangeData[RDS__COMMENT], COMMENT_LEN - 1, szBuffer)

	new iFile = fopen(g_szIpListFile, "a")

	if(!iFile) {
		console_print(pPlayer, "* Error, can't open subnet list file!")
		return PLUGIN_HANDLED
	}

	fprintf(iFile, "^n%s %s %s", LIST_NAME[iType], szStart, szEnd)

	if(szBuffer[0]) {
		fprintf(iFile, " ^"%s^"", szBuffer)
	}

	fclose(iFile)

	ArrayPushArray(g_eIpArray[iType], g_eRangeData)
	g_iIpCount[iType]++

	console_print(pPlayer, "* Range '%s - %s' successfully added to %s!", szStart, szEnd, LIST_NAME[iType])

	log_to_file(g_eLogFile[LOG__CMD], "Range '%s - %s' added to %s by %N", szStart, szEnd, LIST_NAME[iType], pPlayer)

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_DelFromIpList(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	new szCmd[64], iType

	read_argv(0, szCmd, chx(szCmd))

	if(equali(szCmd, "bg_ip_blacklist_del")) {
		iType = LIST_TYPE__BLACKLIST
	}
	else {
		iType = LIST_TYPE__WHITELIST
	}

	if(read_argc() < 3) {
		console_print( pPlayer, "* Usage: %s <start_ip> <end_ip>",
			iType == LIST_TYPE__BLACKLIST ? "bg_ip_blacklist_del" : "bg_ip_whitelist_del" );

		return PLUGIN_HANDLED
	}

	new szStart[IP_LEN], szEnd[IP_LEN]

	read_argv(1, szStart, chx(szStart))
	read_argv(2, szEnd, chx(szEnd))

	if(!IsIpValid(szStart) || !IsIpValid(szEnd)) {
		console_print(pPlayer, "* Wrong value, check your input!")
		return PLUGIN_HANDLED
	}

	new iStartIP = func_ParseIP(szStart)
	new iEndIP = func_ParseIP(szEnd)

	new bool:bFound

	for(new i; i < g_iIpCount[iType]; i++) {
		ArrayGetArray(g_eIpArray[iType], i, g_eRangeData)

		if(g_eRangeData[RDS__START_IP] == iStartIP && g_eRangeData[RDS__END_IP] == iEndIP) {
			ArrayDeleteItem(g_eIpArray[iType], i)
			g_iIpCount[iType]--
			bFound = true
			break
		}
	}

	if(!bFound) {
		console_print(pPlayer, "* Record with provided values not found!")
		return PLUGIN_HANDLED
	}

	new szOldFileName[PLATFORM_MAX_PATH]
	formatex(szOldFileName, chx(szOldFileName), "%s_old", g_szIpListFile)

	if(!rename_file(g_szIpListFile, szOldFileName, .relative = 1)) {
		console_print(pPlayer, "* Error, can't rename subnet list file!")
		return PLUGIN_HANDLED
	}

	new iOldFile = fopen(szOldFileName, "r")

	if(!iOldFile) {
		console_print(pPlayer, "* Error, can't open renamed subnet list file!")
		return PLUGIN_HANDLED
	}

	new iNewFile = fopen(g_szIpListFile, "w")

	if(!iNewFile) {
		fclose(iOldFile)
		console_print(pPlayer, "* Error, can't open new subnet list file!")
		return PLUGIN_HANDLED
	}

	func_AddDefSting_List(iNewFile)

	new iCfgType, szBuffer[128], szOldStart[IP_LEN], szOldEnd[IP_LEN], szComment[COMMENT_LEN]

	bFound = false

	while(!feof(iOldFile)) {
		fgets(iOldFile, szBuffer, chx(szBuffer))
		trim(szBuffer)

		switch(szBuffer[0]) {
			case 'b': iCfgType = LIST_TYPE__BLACKLIST
			case 'w': iCfgType = LIST_TYPE__WHITELIST
			default: continue //  ';', '/', EOS, etc.
		}

		szComment[0] = EOS

		parse(szBuffer, "", "", szOldStart, chx(szOldStart), szOldEnd, chx(szOldEnd), szComment, COMMENT_LEN - 1)

		if(equal(szStart, szOldStart) && equal(szEnd, szOldEnd) && iType == iCfgType) {
			bFound = true
			continue
		}

		fprintf(iNewFile, "^n%s %s %s", LIST_NAME[iCfgType], szOldStart, szOldEnd)

		if(szComment[0]) {
			fprintf(iNewFile, " ^"%s^"", szComment)
		}
	}

	fclose(iOldFile)
	fclose(iNewFile)

	if(!delete_file(szOldFileName)) {
		console_print(pPlayer, "* Warning! Can't delete '%s'", szOldFileName)
	}

	if(bFound) {
		console_print(pPlayer, "* Range '%s - %s' successfully removed from %s!", szStart, szEnd, LIST_NAME[iType])
	}
	else {
		console_print(pPlayer, "* Range '%s - %s' removed from %s, but not found in subnet list file!", szStart, szEnd, LIST_NAME[iType])
	}

	log_to_file(g_eLogFile[LOG__CMD], "Range '%s - %s' removed from %s by %N", szStart, szEnd, LIST_NAME[iType], pPlayer)

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_ShowIpList(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	new szBuffer[64], iType

	read_argv(0, szBuffer, chx(szBuffer))

	if(equali(szBuffer, "bg_ip_blacklist_show")) {
		iType = LIST_TYPE__BLACKLIST
	}
	else {
		iType = LIST_TYPE__WHITELIST
	}

	new iTotalPages = floatround(float(g_iIpCount[iType]) / 10.0, floatround_ceil)

	if(read_argc() == 1) {
		console_print( pPlayer, "* Usage: %s <page> (total pages: %d)",
			iType == LIST_TYPE__BLACKLIST ? "bg_ip_blacklist_show" : "bg_ip_whitelist_show", iTotalPages );

		return PLUGIN_HANDLED
	}

	new iPage = read_argv_int(1)

	if(iPage < 1) {
		iPage = 1
	}

	if(iPage > iTotalPages) {
		iPage = iTotalPages
	}

	if(!iPage) {
		console_print(pPlayer, "* IP %s is empty!", LIST_NAME[iType])
		return PLUGIN_HANDLED
	}

	console_print(pPlayer, "Displaying page %d/%d:", iPage, iTotalPages)
	console_print(pPlayer, "# <-> StartIP <-> EndIP <-> Comment")

	new iCount, iStartPos, i
	iStartPos = i = 10 * (iPage - 1)

	new szStartIP[IP_LEN], szEndIP[IP_LEN]

	while(i < g_iIpCount[iType]) {
		ArrayGetArray(g_eIpArray[iType], i, g_eRangeData)

		func_ReverseIP(g_eRangeData[RDS__START_IP], szStartIP, chx(szEndIP))
		func_ReverseIP(g_eRangeData[RDS__END_IP], szEndIP, chx(szEndIP))

		if(!g_eRangeData[RDS__COMMENT][0]) {
			g_eRangeData[RDS__COMMENT] = _NA_
		}

		console_print(pPlayer, "%d <-> %s <-> %s <-> %s", i + 1, szStartIP, szEndIP, g_eRangeData[RDS__COMMENT])

		if(++iCount == 10) {
			break
		}

		i++
	}

	console_print(pPlayer, "Displayed records: %d/%d", iStartPos + iCount, g_iIpCount[iType])

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_FindIP(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(read_argc() == 1) {
		console_print(pPlayer, "* Usage: bg_find_ip <ip> - Check IP presence in blacklist and whitelist")
		return PLUGIN_HANDLED
	}

	new szBuffer[64]

	read_args(szBuffer, chx(szBuffer))
	remove_quotes(szBuffer)
	trim(szBuffer)

	if(!IsIpValid(szBuffer)) {
		console_print(pPlayer, "* Wrong IP, check your input!")
		return PLUGIN_HANDLED
	}

	new iIP = func_ParseIP(szBuffer)
	new szStartIP[IP_LEN], szEndIP[IP_LEN]

	for(new iType = LIST_TYPE__BLACKLIST; iType <= LIST_TYPE__WHITELIST; iType++) {
		if(IsIpInList(iIP, iType)) {
			console_print(pPlayer, "* IP '%s' FOUND in %s", szBuffer, LIST_NAME[iType])

			if(!g_eRangeData[RDS__COMMENT][0]) {
				g_eRangeData[RDS__COMMENT] = _NA_
			}

			func_ReverseIP(g_eRangeData[RDS__START_IP], szStartIP, chx(szStartIP))
			func_ReverseIP(g_eRangeData[RDS__END_IP], szEndIP, chx(szEndIP))

			console_print(pPlayer, "* Range '%s <-> %s', comment: '%s'", szStartIP, szEndIP, g_eRangeData[RDS__COMMENT])
			continue
		}

		console_print(pPlayer, "* IP '%s' NOT found in %s", szBuffer, LIST_NAME[iType])
	}

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_GetAsByIP(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(read_argc() == 1) {
		console_print(pPlayer, "* Usage: bg_get_as_by_ip <ip> - Get AS number for provided IP")
		return PLUGIN_HANDLED
	}

	new szBuffer[64]

	read_args(szBuffer, chx(szBuffer))
	remove_quotes(szBuffer)
	trim(szBuffer)

	if(!IsIpValid(szBuffer)) {
		console_print(pPlayer, "* Wrong IP, check your input!")
		return PLUGIN_HANDLED
	}

	if(TrieGetArray(g_tIpInfo, szBuffer, g_eIpInfoData, sizeof(g_eIpInfoData))) {
		if(!g_eIpInfoData[IP_INFO__DESC][0]) {
			copy(g_eIpInfoData[IP_INFO__DESC], DESC_LEN - 1, _NA_)
		}
	}
	else {
		if(!equal(g_szGetWhoisIP, szBuffer)) {
			g_iManualWhoisState = 0
			copy(g_szGetWhoisIP, chx(g_szGetWhoisIP), szBuffer)
			new eExtData[WHOIS_DATA_STRUCT]; eExtData[WHOIS_DATA__QUERY_TYPE] = WHOIS_QUERY_TYPE__MANUAL
			WhoisIpQuery(szBuffer, "OnWhoisIpQueryComplete", eExtData, sizeof(eExtData))
			console_print(pPlayer, "* Query for IP '%s' was sent, use cmd again!", szBuffer)
			log_to_file(g_eLogFile[LOG__CMD], "GetAsByIp request for '%s' by %N", szBuffer, pPlayer)
			return PLUGIN_HANDLED
		}

		g_szGetWhoisIP[0] = EOS

		if(g_iManualWhoisState == INVALID_HANDLE) {
			console_print(pPlayer, "* Query error occured, try again!")
			return PLUGIN_HANDLED
		}

		if(!g_iManualWhoisState) {
			console_print(pPlayer, "* No data has been recieved! Try again!")
			return PLUGIN_HANDLED
		}
	}

	copy(g_szIP, chx(g_szIP), szBuffer)
	func_GetCountryCode(0)
	func_GetCountryName(0)

	console_print( pPlayer, "* AS number for IP '%s' (%s, %s) is '%s' (%s)", szBuffer, g_szCode[0],
		g_szCountry[0], g_eIpInfoData[IP_INFO__AS], g_eIpInfoData[IP_INFO__DESC] );

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_FindAS(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(read_argc() == 1) {
		console_print(pPlayer, "* Usage: bg_find_as <as_number> - Check AS number presence in blacklist and whitelist")
		return PLUGIN_HANDLED
	}

	new szBuffer[64]

	read_args(szBuffer, chx(szBuffer))
	remove_quotes(szBuffer)
	trim(szBuffer)

	if(!IsValidAsNumber(szBuffer)) {
		console_print(pPlayer, "* Wrong AS number, check your input!")
		return PLUGIN_HANDLED
	}

	if(!TrieKeyExists(g_tAsNumbers, szBuffer)) {
		console_print(pPlayer, "* AS number '%s' NOT found in both lists!", szBuffer)
		return PLUGIN_HANDLED
	}

	for(new iType = LIST_TYPE__BLACKLIST; iType <= LIST_TYPE__WHITELIST; iType++) {
		for(new i; i < g_iAsCount[iType]; i++) {
			ArrayGetArray(g_aAsArray[iType], i, g_eAsData)

			if(equal(g_eAsData[ASD__NUMBER], szBuffer)) {
				if(!g_eAsData[ASD__COMMENT][0]) {
					g_eAsData[ASD__COMMENT] = _NA_
				}

				console_print(pPlayer, "* AS number '%s' FOUND in %s, comment: '%s'", szBuffer, LIST_NAME[iType], g_eAsData[ASD__COMMENT])
				return PLUGIN_HANDLED
			}
		}
	}

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_CheckIP(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(read_argc() == 1) {
		console_print(pPlayer, "* Usage: bg_check_ip <ip> - Check IP address for Proxy/VPN involvement (bypasing all cache)")
		return PLUGIN_HANDLED
	}

	if(!g_bSystemAvailable) {
		console_print(pPlayer, "* System not initialized yet!")
		return PLUGIN_HANDLED
	}

	new szBuffer[64]

	read_args(szBuffer, chx(szBuffer))
	remove_quotes(szBuffer)
	trim(szBuffer)

	if(!IsIpValid(szBuffer)) {
		console_print(pPlayer, "* Wrong IP, check your input!")
		return PLUGIN_HANDLED
	}

	copy(g_szIP, chx(g_szIP), szBuffer) // need for both conditions

	new iIpStatus

	if(!equal(g_szGetProxyIP, szBuffer)) {
		copy(g_szGetProxyIP, chx(g_szGetProxyIP), szBuffer)
		TrieDeleteKey(g_tIpProxyStatus, szBuffer)
		func_CheckForProxy(INVALID_HANDLE)
		console_print(pPlayer, "* Query for IP '%s' was sent, use cmd again!", szBuffer)
		log_to_file(g_eLogFile[LOG__CMD], "CheckIp request for '%s' by %N", szBuffer, pPlayer)
		return PLUGIN_HANDLED
	}

	g_szGetProxyIP[0] = EOS

	if(!TrieGetCell(g_tIpProxyStatus, szBuffer, iIpStatus)) {
		console_print(pPlayer, "* No data has been recieved (error?), try again!")
		return PLUGIN_HANDLED
	}

	func_GetCountryCode(0)
	func_GetCountryName(0)

	console_print( pPlayer, "* '%s' (%s, %s) status is '%s'", szBuffer, g_szCode[0],
		g_szCountry[0], IP_STATES[iIpStatus] );

	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_FlushData(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(read_argc() != 1) {
		log_to_file(g_eLogFile[LOG__CMD], "Flush cmd with param '%d' by %N", read_argv_int(1), pPlayer)

		switch(read_argv_int(1)) {
			case 1: {
				if(!g_bSystemAvailable) {
					console_print(pPlayer, "* System not initialized yet!")
					return PLUGIN_HANDLED
				}

				// TrieClear(g_tIpProxyStatus) -> In SQL_Handler(), see 'QUERY__FLUSH_IPS'
				formatex(g_szQuery, chx(g_szQuery), "DELETE FROM `%s`", g_szTableName)
				func_MakeQuery(QUERY__FLUSH_IPS)
				console_print(pPlayer, "* Proxy/VPN IP check data flushed!")
				return PLUGIN_HANDLED
			}
			case 2: {
				nvault_prune(g_hImmunity, 0, get_systime())
				console_print(pPlayer, "* SteamID immunity data flushed!")
				return PLUGIN_HANDLED
			}
			case 3: {
				TrieClear(g_tAsCache)
				console_print(pPlayer, "* AS check status cache flushed!")
				return PLUGIN_HANDLED
			}
			case 4: {
				g_iIpCount[LIST_TYPE__BLACKLIST] = 0
				g_iIpCount[LIST_TYPE__WHITELIST] = 0
				ArrayClear(Array:g_eIpArray[LIST_TYPE__BLACKLIST])
				ArrayClear(Array:g_eIpArray[LIST_TYPE__WHITELIST])

				new iFile = fopen(g_szIpListFile, "w")

				if(iFile) {
					func_AddDefSting_List(iFile)
					console_print(pPlayer, "* IP range configuration flushed!")
					fclose(iFile)
				}
				else {
					console_print(pPlayer, "* Can't open '%s'!", IP_FILE_NAME)
				}

				return PLUGIN_HANDLED
			}
			case 5: {
				g_iAsCount[LIST_TYPE__BLACKLIST] = 0
				g_iAsCount[LIST_TYPE__WHITELIST] = 0
				TrieClear(g_tAsCache)
				TrieClear(g_tAsNumbers)
				ArrayClear(Array:g_aAsArray[LIST_TYPE__BLACKLIST])
				ArrayClear(Array:g_aAsArray[LIST_TYPE__WHITELIST])

				new iFile = fopen(g_szAsListFile, "w")

				if(iFile) {
					func_AddDefSting_AS(iFile)
					console_print(pPlayer, "* AS configuraion and check status cache flushed!")
					fclose(iFile)
				}
				else {
					console_print(pPlayer, "* Can't truncate '%s'!", AS_FILE_NAME)
				}

				return PLUGIN_HANDLED
			}
		}
	}

	console_print(pPlayer, "* Usage: bg_flush_data <1-5>")
	console_print(pPlayer, "* 1 - Proxy/VPN IP check status (sql + trie)")
	console_print(pPlayer, "* 2 - SteamID immunity list (nvault)")
	console_print(pPlayer, "* 3 - AS check status cache (trie)")
	console_print(pPlayer, "* 4 - IP range configuration (memory + file)")
	console_print(pPlayer, "* 5 - AS configuration (memory + file)")
	return PLUGIN_HANDLED
}

/* -------------------- */

public concmd_Status(pPlayer, iAccess) {
	if(!UserHasAccess(pPlayer, iAccess)) {
		return PLUGIN_HANDLED
	}

	if(!g_eCvar[CVAR__PLUGIN_ENABLED]) {
		console_print(pPlayer, "* Plugin disabled, command unavailable!")
		return PLUGIN_HANDLED
	}

	console_print(pPlayer, "^nPlayers status:^n# <-> Access <-> Nick <-> IP <-> SteamID <-> AS <-> Desc. <-> Code <-> Country")

	new iCount

	for(new pUser = 1; pUser <= MaxClients; pUser++) {
		if(!is_user_connected(pUser)) {
			continue
		}

		get_user_ip(pUser, g_szIP, chx(g_szIP), .without_port = 1)
		get_user_authid(pUser, g_szAuthID, chx(g_szAuthID))

		console_print( pPlayer, "%d <-> %s <-> %n <-> %s <-> %s <-> %s <-> %s <-> %s <-> %s", ++iCount, g_szAccess[pUser],
			pUser, g_szIP, g_szAuthID, g_szAsNumber[pUser], g_szDesc[pUser], g_szCode[pUser], g_szCountry[pUser] );
	}

	console_print(pPlayer, "Total: %d^n", iCount)

	return PLUGIN_HANDLED
}

/* -------------------- */

stock func_LoadRanges() {
	new iFile = fopen(g_szIpListFile, "r")

	if(!iFile) {
		if(file_exists(g_szIpListFile)) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open existing '%s'", IP_FILE_NAME)
			return
		}

		iFile = fopen(g_szIpListFile, "w")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't create default '%s'", IP_FILE_NAME)
			return
		}

		func_AddDefSting_List(iFile)
		fclose(iFile)
		return
	}

	new iType, szStartIP[IP_LEN], szEndIP[IP_LEN], szBuffer[128]

	while(!feof(iFile)) {
		fgets(iFile, szBuffer, chx(szBuffer))
		trim(szBuffer)

		switch(szBuffer[0]) {
			case 'b': iType = LIST_TYPE__BLACKLIST
			case 'w': iType = LIST_TYPE__WHITELIST
			default: continue //  ';', '/', EOS, etc.
		}

		g_eRangeData[RDS__COMMENT][0] = EOS

		parse(szBuffer, "", "", szStartIP, chx(szStartIP), szEndIP, chx(szEndIP), g_eRangeData[RDS__COMMENT], COMMENT_LEN - 1)

		g_eRangeData[RDS__START_IP] = func_ParseIP(szStartIP)
		g_eRangeData[RDS__END_IP] = func_ParseIP(szEndIP)

		ArrayPushArray(g_eIpArray[iType], g_eRangeData)
		g_iIpCount[iType]++
	}

	fclose(iFile)
}

/* -------------------- */

public func_LoadAsNumbers() {
	new iFile = fopen(g_szAsListFile, "r")

	if(!iFile) {
		if(file_exists(g_szAsListFile)) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open existing '%s'", AS_FILE_NAME)
			return
		}

		iFile = fopen(g_szAsListFile, "w")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't create default '%s'", AS_FILE_NAME)
			return
		}

		func_AddDefSting_AS(iFile)
		fclose(iFile)
		return
	}

	new iType, szBuffer[128]

	while(!feof(iFile)) {
		fgets(iFile, szBuffer, chx(szBuffer))
		trim(szBuffer)

		switch(szBuffer[0]) {
			case 'b': iType = LIST_TYPE__BLACKLIST
			case 'w': iType = LIST_TYPE__WHITELIST
			default: continue //  ';', '/', EOS, etc.
		}

		g_eAsData[ASD__COMMENT][0] = EOS

		parse(szBuffer, "", "", g_eAsData[ASD__NUMBER], AS_LEN - 1, g_eAsData[ASD__COMMENT], COMMENT_LEN - 1)

		g_eAsTrieData[AST__STATE] = (iType == LIST_TYPE__BLACKLIST) ? STATE__BAN : STATE__WHITELIST;
		g_eAsTrieData[AST__ARRAY_POS] = g_iAsCount[iType]
		TrieSetArray(g_tAsNumbers, g_eAsData[ASD__NUMBER], g_eAsTrieData, sizeof(g_eAsTrieData))

		ArrayPushArray(g_aAsArray[iType], g_eAsData)
		g_iAsCount[iType]++
	}

	fclose(iFile)
}

/* -------------------- */

stock func_LoadAllowedCodes(szCfgDir[]) {
	new iFile = fopen(szCfgDir, "r")

	if(!iFile) {
		if(file_exists(szCfgDir)) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open existing '%s'", COUNTRY_FILE_NAME)
			return
		}

		iFile = fopen(szCfgDir, "w")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't create default '%s'", COUNTRY_FILE_NAME)
			return
		}

		fputs( iFile,
			"; Список стран, которые могут входить на сервер^n\
			; Country list that can enter to the server^n\
			^"RU^" // Россия^n\
			^"UA^" // Украина^n\
			^"BY^" // Беларусь^n\
			^"KZ^" // Казахстан^n\
			^"KG^" // Киргизия^n\
			^"MD^" // Молдавия^n\
			^"UZ^" // Узбекистан^n\
			^"LV^" // Латвия^n\
			^"LT^" // Литва^n\
			^"EE^" // Эстония^n\
			^"GE^" // Грузия^n\
			^"RO^" // Румыния^n\
			^"AM^" // Армения^n\
			^"BG^" // Болгария^n\
			^"AZ^" // Азербайджан^n\
			^"TM^" // Туркменистан^n\
			^"TJ^" // Таджикистан"
		);

		fclose(iFile)

		iFile = fopen(szCfgDir, "r")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open existing '%s'", COUNTRY_FILE_NAME)
			return
		}
	}

	new szBuffer[32]

	while(!feof(iFile)) {
		if(g_iCodesCount == MAX_CODES) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Codes limit reached! Increase 'MAX_CODES' in .sma !")
			break
		}

		fgets(iFile, szBuffer, chx(szBuffer))

		if(szBuffer[0] != '"') {
			continue
		}

		parse(szBuffer, g_szAllowedCodes[g_iCodesCount++], CODE_LEN - 1)
	}

	fclose(iFile)
}

/* -------------------- */

stock func_LoadApiKeys(szKeyFile[]) {
	new iFile = fopen(szKeyFile, "r")

	if(!iFile) {
		if(file_exists(szKeyFile)) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't open existing '%s'", KEY_FILE_NAME)
			return
		}

		iFile = fopen(szKeyFile, "w")

		if(!iFile) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Can't create default '%s'", KEY_FILE_NAME)
			return
		}

		fputs( iFile,
			"; Ключи к API iphub'а. Зарегистрируйтесь на http://iphub.info/ , получите бесплатный ключ (1000 проверок/сутки), и введите его сюда^n\
			; API-keys for iphub. Register at http://iphub.info/ , get free key (1000 checks/day), and set it here^n\
			X-Key: ########################################"
		);

		fclose(iFile)
		return
	}

	new szBuffer[API_KEY_LEN + 8]

	while(!feof(iFile)) {
		if(g_iLoadedKeys == MAX_KEYS) {
			log_to_file(g_eLogFile[LOG__ERROR], "[Error] Keys limit reached! Increase 'MAX_KEYS' in .sma !")
			break
		}

		fgets(iFile, szBuffer, chx(szBuffer))
		trim(szBuffer)

		if(szBuffer[0] != 'X') {
			continue
		}

		copy(g_szApiKey[g_iLoadedKeys++], API_KEY_LEN - 1, szBuffer)
	}

	fclose(iFile)
}

/* -------------------- */

stock func_ParseIP(szIP[]) {
	new szRight[IP_LEN], szPart[4], iIP, iOctet

	strtok(szIP, szPart, chx(szPart), szRight, chx(szRight), '.')

	for(new i; i < 4; i++) {
		iOctet = str_to_num(szPart)

		if(iOctet < 0) {
			iOctet = 0
		}
		else if(iOctet > 255) {
			iOctet = 255
		}

		iIP += iOctet

		if(i == 3) {
			break
		}

		strtok(szRight, szPart, chx(szPart), szRight, chx(szRight), '.')
		iIP = iIP << 8
	}

	return iIP
}

/* -------------------- */

stock func_ReverseIP(iIP, szIP[IP_LEN], iLen) {
	new iOctet[4], bool:bHigh

	if(iIP < 0) {
		bHigh = true;
		iIP = iIP & (~(1 << 31));
	}

	for(new i = 0; i < 4; i++) {
		iOctet[i] = iIP & 255;
		iIP = iIP >> 8;
	}
	if(bHigh) {
		iOctet[3] += 128
	}

	formatex(szIP, iLen, "%d.%d.%d.%d", iOctet[3], iOctet[2], iOctet[1], iOctet[0])
}

/* -------------------- */

stock bool:IsIpInList(iIP, iType) {

	for(new i; i < g_iIpCount[iType]; i++) {
		ArrayGetArray(g_eIpArray[iType], i, g_eRangeData)

		if(CompareUnsigned(g_eRangeData[RDS__START_IP], iIP) <= 0 && CompareUnsigned(g_eRangeData[RDS__END_IP], iIP) >= 0) {
			return true
		}
	}

	return false
}

/* -------------------- */

/**
 * Compares two integers as unsigned values.
 *
 * @param	first	First value to compare.
 * @param	second	Second value to compare.
 * @return			-1 if first is smaller than second.
 *					 0 if first is equal to second.
 *					 1 if first is greater than second.
 */
stock CompareUnsigned(first, second) {
	if(first == second) {
		return 0
	}

	new bool:highFirst, bool:highSecond

	if(first < 0) {
		highFirst = true
		first = first & (~(1 << 31))
	}

	if(second < 0) {
		highSecond = true
		second = second & (~(1 << 31))
	}

	if(highFirst && !highSecond) {
		return 1
	}

	if(!highFirst && highSecond) {
		return -1
	}

	if(first > second) {
		return 1
	}

	return -1
}

/* -------------------- */

public hook_CvarChange(pCvar, szOldVal[], szNewVal[]) {
	g_iDefAccFlags = read_flags(szNewVal)
}

/* -------------------- */

stock bool:IsIpValid(szIP[]) {
	new i, szRight[IP_LEN], szPart[4], iCount

	strtok(szIP, szPart, chx(szPart), szRight, chx(szRight), '.')

	while(szPart[0] >= '0' && szPart[0] <= '9')	{
		i = str_to_num(szPart)

		if(i < 0 || i > 255) {
			return false
		}

		iCount++
		strtok(szRight, szPart, chx(szPart), szRight, chx(szRight), '.')
	}

	return (iCount == 4)
}

/* -------------------- */

stock bool:IsValidSteamID(szAuthID[]) {
	// STEAM_ or VALVE_
	// 0:(0|1):\d+
	// 0-5 = STEAM_
	// 6 = integer
	// 7 = :
	// 8 = integer
	// 9 = :
	// 10+ = integer
	return (
		( equal(szAuthID, "STEAM_", 6) || equal(szAuthID, "VALVE_", 6) )
			&&
		isdigit(szAuthID[6])
			&&
		szAuthID[7] == ':'
			&&
		isdigit(szAuthID[8])
			&&
		szAuthID[9] == ':'
			&&
		is_str_num(szAuthID[10])
	);
}

/* -------------------- */

stock bool:IsValidAsNumber(szAsNumber[]) {
	//As we use trie, we need case-sensitive comparement
	//if((szAsNumber[0] != 'A' && szAsNumber[0] != 'a') || (szAsNumber[1] != 'S' && szAsNumber[1] != 's') || !szAsNumber[2]) {
	if(szAsNumber[0] != 'A' || szAsNumber[1] != 'S' || !szAsNumber[2]) {
		return false
	}

	for(new i = 2; i < AS_LEN; i++) {
		if(!szAsNumber[i]) {
			return true
		}

		if(szAsNumber[i] < '0' || szAsNumber[i] > '9') {
			return false
		}
	}

	return false
}

/* -------------------- */

stock func_AddDefSting_List(iFile) {
	fputs(iFile, ";Line format: blacklist/whitelist start_ip end_ip ^"comment^"[optional]^n")
}

/* -------------------- */

stock func_AddDefSting_AS(iFile) {
	fputs(iFile, ";Line format: blacklist/whitelist AS_number ^"comment^"[optional]^n")
}

/* -------------------- */

stock func_GetCountryCode(pPlayer) {
	new szCode[CODE_LEN]

	if(geoip_code2_ex(g_szIP, szCode))	{
		copy(g_szCode[pPlayer], chx(g_szCode[]), szCode)
		return
	}

	g_szCode[pPlayer] = _NA_
}

/* -------------------- */

stock func_GetCountryName(pPlayer) {
	if(!geoip_country_ex(g_szIP, g_szCountry[pPlayer], COUNTRY_LEN - 1)) {
		g_szCountry[pPlayer] = _NA_
	}
}

/* -------------------- */

stock func_LogAllow(pPlayer) {
	log_to_file( g_eLogFile[LOG__ALLOW], "[%s] %n | %s | %s | %s | %s | %s | %s",
		g_szAccess[pPlayer], pPlayer, g_szIP, g_szAuthID, g_szAsNumber[pPlayer], g_szDesc[pPlayer], g_szCode[pPlayer], g_szCountry[pPlayer] );
}

/* -------------------- */

stock bool:UserHasAccess(pPlayer, iAccess) {
	if(get_user_flags(pPlayer) & iAccess) {
		return true
	}

	console_print(pPlayer, "* You have no access to this command!")
	return false
}

/* -------------------- */

stock func_MakeQuery(iQueryType) {
	g_eSqlData[SQL_DATA__QUERY_TYPE] = iQueryType
	SQL_ThreadQuery(g_hSqlTuple, "SQL_Handler", g_szQuery, g_eSqlData, sizeof(g_eSqlData))
}

/* -------------------- */

public plugin_end() {
	g_bPluginEnded = true

	if(g_hSqlTuple) {
		SQL_FreeHandle(g_hSqlTuple)
	}

	if(g_hImmunity != INVALID_HANDLE) {
		nvault_close(g_hImmunity)
		ArrayDestroy(Array:g_eIpArray[LIST_TYPE__BLACKLIST])
		ArrayDestroy(Array:g_eIpArray[LIST_TYPE__WHITELIST])
		TrieDestroy(g_tAsCache)
		TrieDestroy(g_tAsNumbers)
		ArrayDestroy(Array:g_aAsArray[LIST_TYPE__BLACKLIST])
		ArrayDestroy(Array:g_aAsArray[LIST_TYPE__WHITELIST])
		TrieDestroy(g_tCurl)
		TrieDestroy(g_tIpInfo)
		TrieDestroy(g_tIpProxyStatus)
		TrieDestroy(g_tIpMatch)
	#if defined DEBUG
		nvault_close(g_hDebug)
	#endif
	}
}